import copy
import numpy as np


class State(object):
    def __init__(self,
                 label,
                 energy,
                 multiplicity=1,
                 size=1,
                 molecules_list=None,
                 connected_distance=1):

        self._label = label
        self._energy = energy
        self._multiplicity = multiplicity
        self._size = size
        self._connected_distance = connected_distance
        self._molecules_set = molecules_list if molecules_list is not None else []
        self._cell_state = None
        self.supercell = None

    def __hash__(self):
        return hash((self._label,
                     self._energy,
                     self._multiplicity,
                     self._size,
                     str(self._cell_state) if self._label != ground_state.label else '',
                     tuple(self.get_molecules())))

    def __eq__(self, other):
        return hash(self) == hash(other)

    def copy(self):
        return copy.deepcopy(self)

    def get_molecules(self):
        assert self._molecules_set is not None
        return self._molecules_set

    def get_center(self):
        assert self._molecules_set is not None
        return self._molecules_set[0]

    def get_coordinates(self):
        # print(self.get_center().get_coordinates())
        return np.average([mol.get_coordinates() for mol in self.get_molecules()], axis=0)

    def get_coordinates_relative(self, supercell):

        supercell = np.array(supercell)
        coor_average = []
        for mol in self.get_molecules():
            coor_average.append(mol.get_coordinates() - np.dot(supercell.T, mol.cell_state))

        return np.average(coor_average, axis=0)

    def get_coordinates_absolute(self):

        supercell = np.array(self.supercell)
        coor_average = []
        for mol in self.get_molecules():
            coor_average.append(mol.get_coordinates() - np.dot(supercell.T, mol.cell_state + self.cell_state))

        return np.average(coor_average, axis=0)

    def add_molecule(self, molecule):
        if not molecule in self._molecules_set:
            self._molecules_set.append(molecule)
            if self._cell_state is None:
                self._cell_state = np.zeros_like(molecule.get_coordinates(), dtype=int)

    def reorganize_cell(self):

        cell_diff = np.array(np.average([np.array(mol.cell_state) for mol in self.get_molecules()], axis=0), dtype=int)

        for mol in self.get_molecules():
            mol.cell_state = np.array(mol.cell_state) - cell_diff

        self.cell_state += cell_diff

    def remove_molecules(self):
        self._molecules_set = []
        # self._cell_state = None

    def reset_molecules(self):
        for mol in self.get_molecules():
            # mol.set_state(self)
            mol.cell_state *= 0

    @property
    def label(self):
        return self._label

    @property
    def energy(self):
        return self._energy

    @property
    def multiplicity(self):
        return self._multiplicity

    @property
    def size(self):
        return self._size

    @property
    def connected_distance(self):
        return self._connected_distance

    @property
    def cell_state(self):
        assert self._cell_state is not None
        return self._cell_state

    @cell_state.setter
    def cell_state(self, c_state):
        self._cell_state = np.array(c_state)

    @property
    def vdw_radius(self):
        return np.average([mol.get_vdw_radius() for mol in self.get_molecules()])

ground_state = State(label='gs', energy=0.0, multiplicity=1)


if __name__ == '__main__':
    print('Test state')
    s = State(label='s1', energy=1.5, multiplicity=1)
    print(hash(s))
    s2 = State(label='s1', energy=1.5, multiplicity=1)
    s2.cell_state = [1, 2]
    print(hash(s2))

    print(s == s2)
    print(s in [s])
    print(s2 in [s])


import numpy as np
from kimonet.utils.units import BOLTZMANN_CONSTANT
from scipy.integrate import quad
import math

###########################################################################################################
#                                 Frank-Condon weighted density
###########################################################################################################
overlap_data = {}


# deprecated
def general_fcwd(donor, acceptor, process, conditions):
    """
    :param donor:
    :param acceptor:
    :param process:
    :param conditions:
    :return: The spectral overlap between the donor and the acceptor
    """

    # testing normalization

    # info = str(hash((donor, acceptor, process, str(conditions), 'general_fcwd')))
    # info = str(hash(donor) + hash(acceptor) + hash(str(conditions) + 'general_fcwd'))

    transition_donor = (process.initial[0], process.final[0])
    transition_acceptor = (process.initial[1], process.final[1])

    donor_vib_dos = donor.get_vib_dos(transition_donor)
    acceptor_vib_dos = acceptor.get_vib_dos(transition_acceptor)

    # print(donor_vib_dos)
    info = str(hash(donor_vib_dos) + hash(acceptor_vib_dos))

    if info in overlap_data:
        # the memory is used if the overlap has been already computed
        return overlap_data[info]

    # test_donor = quad(donor_vib_dos, 0, np.inf,  epsabs=1e-20)[0]
    # test_acceptor = quad(acceptor_vib_dos, 0, np.inf,  epsabs=1e-20)[0]

    # print('test_donor', test_donor)
    # print('test_acceptor', test_acceptor)

    # assert math.isclose(test_donor, 1.0, abs_tol=0.01)
    # assert math.isclose(test_acceptor, 1.0, abs_tol=0.01)

    def overlap(x):
        return donor_vib_dos(x) * acceptor_vib_dos(x)

    overlap_data[info] = quad(overlap, 0, np.inf,  epsabs=1e-5, limit=1000)[0]

    return overlap_data[info]

    # return quad(integrand, 0, np.inf, args=(donor, acceptor))[0]


# deprecated (only used in test)
def marcus_fcwd_old(donor, acceptor, conditions):
    """
    :param donor:
    :param acceptor:
    :param conditions:
    :return: The spectral overlap between the donor and the acceptor according to Marcus formula.
    """
    import warnings

    warnings.warn("This method will be deprecated, use vibrations classes instead",
                  DeprecationWarning)

    T = conditions['temperature']       # temperature (K)

    excited_state = donor.state.label
    gibbs_energy = donor.state_energies[excited_state] - acceptor.state_energies[excited_state]
    # Gibbs energy: energy difference between the equilibrium points of the excited states

    reorganization = acceptor.reorganization_energies[('gs', 's1')] + acceptor.reorganization_energies[('s1', 'gs')]
    # acceptor reorganization energy of the excited state

    info = str(hash((T, gibbs_energy, reorganization, 'marcus')))
    # we define a compact string with the characteristic information of the spectral overlap

    if info in overlap_data:
        # the memory is used if the overlap has been already computed
        overlap = overlap_data[info]

    else:
        overlap = 1.0 / (2 * np.sqrt(np.pi * BOLTZMANN_CONSTANT * T * reorganization)) * \
                  np.exp(-(gibbs_energy+reorganization)**2 / (4 * BOLTZMANN_CONSTANT * T * reorganization))

        overlap_data[info] = overlap
        # new values are added to the memory

    return overlap
    # Since we have a quantity in 1/eV, we use the converse function from_ev_to_au in inverse mode
    # to have a 1/au quantity.



import numpy as np
from kimonet.utils.rotation import rotate_vector
import itertools
import warnings


def minimum_distance_vector(r_vector, supercell):
    warnings.warn("deprecated", DeprecationWarning)
    # lattice periodicity
    r_vector = np.array(r_vector, dtype=float).copy()
    cell_vector = []

    for lattice in supercell:
        half_cell = np.array(lattice) / 2
        dot_ref = np.dot(half_cell, half_cell)
        dot = np.dot(half_cell, r_vector)
        n_n = (np.sqrt(np.abs(dot)) // np.sqrt(dot_ref)) * np.sign(dot)
        r_vector += np.array(lattice) * -n_n
        cell_vector.append(-n_n)

    return r_vector, np.array(cell_vector, dtype=int)


def get_supercell_increments(supercell, radius):
    # TODO: This function can be optimized as a function of the particular molecule coordinates
    v = np.array(radius / np.linalg.norm(supercell, axis=1), dtype=int) + 1  # here extensive approximation
    return list(itertools.product(*[range(-i, i + 1) for i in v]))


def distance_vector_periodic(r, supercell, cell_increment):
    return r + np.dot(cell_increment, supercell)


def old_distance_between_molecules(molecule1, molecule2, supercell, cell_state1, cell_state2):
    warnings.warn("deprecated, use get_coordinates_absolute() method", DeprecationWarning)
    r_vector = molecule2.get_coordinates() - molecule1.get_coordinates()
    cell_incr = np.array(cell_state2 - cell_state1)
    return np.linalg.norm(distance_vector_periodic(r_vector, supercell, cell_incr))





class Transition:
    def __init__(self, state1, state2, symmetric=True, tdm=0, reorganization_energy=0):
        self._state1 = state1
        self._state2 = state2
        self._symmetric = symmetric
        self._tdm = tdm
        self._reorganization_energy = reorganization_energy

    def __str__(self):
        return '{} -> {}'.format(self._state1.label, self._state2.label)

    def __hash__(self):
        if self._symmetric:
            return hash(self._state1.label) + hash(self._state2.label)
        else:
            return hash((self._state1.label, self._state2.label))

    def __eq__(self, other):
        if (self._state1.label, self._state2.label) == (other._state1.label, other._state2.label):
            return True
        if self._symmetric or other._symmetric:
            if (self._state1.label, self._state2.label) == (other._state2.label, other._state1.label):
                return True

        return hash(self) == hash(other)

    @property
    def tdm(self):
        return self._tdm

    @property
    def reorganization_energy(self):
        return self._reorganization_energy


import numpy as np
import itertools
from copy import deepcopy
from kimonet.utils import distance_vector_periodic
from kimonet.utils import get_supercell_increments
from kimonet.system.state import ground_state as _GS_

combinations_data = {}


def distance_matrix(molecules_list, supercell, max_distance=1):

    # Set maximum possible distance everywhere
    distances = np.ones((len(molecules_list), len(molecules_list))) * np.product(np.diag(supercell))

    for i, mol_i in enumerate(molecules_list):
        for j, mol_j in enumerate(molecules_list):

            coordinates = mol_i.get_coordinates()
            center_position = mol_j.get_coordinates()
            cell_increments = get_supercell_increments(supercell, max_distance)
            for cell_increment in cell_increments:
                r_vec = distance_vector_periodic(coordinates - center_position, supercell, cell_increment)
                d = np.linalg.norm(r_vec)
                if distances[i, j] > d:
                    distances[i, j] = d

    return distances


def is_connected(molecules_list, supercell, connected_distance=1):

    # check if connected
    for group in molecules_list:
        d = distance_matrix(group, supercell, max_distance=connected_distance)
        if len(d) > 1 and  not np.all(connected_distance >= np.min(np.ma.masked_where(d == 0, d), axis=1)):
            return False

    return True


def is_connected_states(molecules_list, supercell, final_states):

    # check if connected
    for i, group in enumerate(molecules_list):
        connected_distance = final_states[i].connected_distance
        d = distance_matrix(group, supercell, max_distance=connected_distance)
        if len(d) > 1 and not np.all(connected_distance >= np.min(np.ma.masked_where(d == 0, d), axis=1)):
            return False

    return True


def partition(elements_list, group_list):

    partition = []
    n = 0
    for i in group_list:
        partition.append(tuple(elements_list[n: n+i]))
        n += i

    return partition


def combinations_group(elements_list_o, process_final, supercell=None, include_self=True, ref_states=None):

    group_list = tuple([s.size for s in process_final])

    elements_list = tuple(range(len(elements_list_o)))
    combinations_list = []

    if (elements_list, group_list) not in combinations_data:

        def combination(data_list, i, cumulative_list):
            if i == 0:
                cumulative_list = []

            for data in itertools.combinations(data_list, group_list[i]):
                rest = tuple(set(data_list) - set(data))
                cumulative_list = cumulative_list[:i] + [data]
                if len(rest) > 0:
                    combination(rest, i+1, deepcopy(cumulative_list))
                else:
                    combinations_list.append(cumulative_list)

        combination(elements_list, 0, [])
        combinations_data[(elements_list, group_list)] = combinations_list
    else:
        combinations_list = combinations_data[(elements_list, group_list)]

    # filter by connectivity
    combinations_list = [[[elements_list_o[l] for l in state] for state in conf] for conf in combinations_list]
    if supercell is not None:
        for c in combinations_list:
            if not is_connected_states(c, supercell, process_final):
                combinations_list.remove(c)

    if not include_self:
        combinations_list = combinations_list[1:]

    return combinations_list


def get_molecules_centered_in_mol(central_mol, molecules_list, supercell, size=1, connected_distance=2):
    for c in itertools.combinations(molecules_list, size):
        if central_mol in c and is_connected([c], supercell, connected_distance=connected_distance):
            return list(c)

    return None


if __name__ == '__main__':

    from kimonet.system.state import State
    from kimonet.system.molecule import Molecule

    test_list = ['a', 'b', 'c']

    group_list = [1, 2]
    configuration = partition(test_list, group_list)
    print(configuration)

    combinations_list = combinations_group(test_list, group_list)

    for c in combinations_list:
        print(c, '---', configuration)
        print(c == configuration)
        print(c)

    s1 = State(label='s1', energy=1.0, multiplicity=1)
    molecule1 = Molecule(coordinates=[0])
    molecule2 = Molecule(coordinates=[1])
    molecule3 = Molecule(coordinates=[2])
    molecule4 = Molecule(coordinates=[4])

    supercell = [[6]]

    test_list = [molecule1, molecule2, molecule3, molecule4]

    group_list = [1, 3]

    configuration = partition(test_list, group_list)

    print('initial:', test_list)
    combinations_list = combinations_group(test_list, group_list, supercell)

    for c in combinations_list:
        print('total: ', c)

    a = distance_matrix(test_list, supercell, max_distance=1)
    print(a)

    print('centered:', molecule1)
    mols = get_molecules_centered_in_mol(molecule1, test_list, supercell, size=3)
    print('mols: ', mols)
    exit()



import numpy as np
import itertools
from kimonet.system import System


def regular_ordered_system(molecule,
                           lattice=None,
                           orientation=None,
                           ):

    if lattice is None:
        lattice = {'size': [1], 'parameters': [1.0]}  # default 1D 1 molecule system

    molecules = []                              # list of instances of class molecule
    for subset in itertools.product(*[list(range(n)) for n in lattice['size']]):
        coordinates = np.multiply(subset, lattice['parameters'])

        molecule = molecule.copy()  # copy of the generic instance
        molecule.set_coordinates(coordinates)
        molecule.set_orientation(orientation)
        molecules.append(molecule)

    supercell = np.diag(np.multiply(lattice['size'], lattice['parameters']))

    return System(molecules, supercell)


def regular_system(molecule,
                   lattice=None,
                   orientation=None,
                   ):

    if lattice is None:
        lattice = {'size': [1], 'parameters': [1.0]}  # default 1D 1 molecule system

    molecules = []  # list of instances of class molecule
    for subset in itertools.product(*[list(range(n)) for n in lattice['size']]):
        coordinates = np.multiply(subset, lattice['parameters'])

        if orientation is None:
            orientation = np.random.random_sample(3) * 2*np.pi

        molecule = molecule.copy()  # copy of the generic instance
        molecule.set_coordinates(coordinates)
        molecule.set_orientation(orientation)
        molecules.append(molecule)

    supercell = np.diag(np.multiply(lattice['size'], lattice['parameters']))

    return System(molecules, supercell)


def crystal_system(molecules,
                   scaled_site_coordinates,
                   dimensions=None,
                   unitcell=None,
                   orientations=None,
                   ):

    unitcell = np.array(unitcell)
    scaled_site_coordinates = np.array(scaled_site_coordinates)
    n_mol, n_dim = scaled_site_coordinates.shape

    if dimensions is None:
        dimensions = [1] * n_dim

    if orientations is None:
        orientations = [None for _ in range(n_mol)]

    molecules_list = []                              # list of instances of class molecule

    for i, (coordinate, molecule) in enumerate(zip(scaled_site_coordinates, molecules)):
        for subset in itertools.product(*[list(range(n)) for n in dimensions]):

            r_cell = np.sum([s * lattice_vector for s, lattice_vector in zip(subset, unitcell)], axis=0)
            coor = r_cell + np.dot(unitcell.T, coordinate)

            molecule = molecule.copy()  # copy of the generic instance
            molecule.set_coordinates(coor)
            molecule.name = 'a{}'.format(i+1)

            if orientations[i] is None:
                final_orientation = np.random.random_sample(3) * 2 * np.pi
            else:
                final_orientation = orientations[i]

            molecule.set_orientation(final_orientation)
            molecules_list.append(molecule)

    supercell = np.dot(unitcell.T, np.diag(dimensions)).T

    return System(molecules_list, supercell)



__version__ = '0.1.1'
from kimonet.core import do_simulation_step, system_test_info
from kimonet.analysis import Trajectory, visualize_system
from warnings import warn
import numpy as np
import time


def calculate_kmc(system, num_trajectories=100, max_steps=10000, silent=False):

    trajectories = []
    for j in range(num_trajectories):
        system_copy = system.copy()

        if not silent:
            print('Trajectory: {}'.format(j))

        trajectory = Trajectory(system_copy)

        for i in range(max_steps):

            change_step, step_time = do_simulation_step(system_copy)
            if system_copy.is_finished:
                break

            trajectory.add_step(change_step, step_time)

            #print('===============================')
            #system_test_info(system_copy)
            #visualize_system(system_copy)

            if i == max_steps-1:
                warn('Maximum number of steps reached!!')

        # trajectory.plot_2d('s1').show()
        # trajectory.plot_2d('t1').show()
        # trajectory.plot_distances('s1').show()
        # trajectory.plot_distances('s2').show()
        # trajectory.plot_graph().show()
        #for node in trajectory.graph.nodes:
        #    print(trajectory.graph.nodes[node]['state'])
        trajectories.append(trajectory)

    return trajectories


def _run_trajectory(index, system, max_steps, silent):
    np.random.seed(int(index * time.time() % 1 * 1e8))

    system_copy = system.copy()
    trajectory = Trajectory(system_copy)
    for i in range(max_steps):

        chosen_process, step_time = do_simulation_step(system_copy)
        if system_copy.is_finished:
            break

        trajectory.add_step(chosen_process, step_time)

        if i == max_steps-1:
            warn('Maximum number of steps reached!!')

    if not silent:
        print('Trajectory {} done!'.format(index))
    return trajectory



def _run_trajectory_intermediate(args):
    system, max_steps, silent = args[1]
    index = args[0]
    return _run_trajectory(index, system, max_steps, silent)


def calculate_kmc_parallel_py2(system, num_trajectories=100, max_steps=10000, silent=False, processors=2):
    import itertools
    from multiprocessing import Pool #, freeze_support

    pool = Pool(processes=processors)
    a_args = list(range(num_trajectories))
    second_arg = [system, max_steps, silent]
    trajectories = pool.map(_run_trajectory_intermediate, itertools.izip(a_args, itertools.repeat(second_arg)))
    return trajectories


def calculate_kmc_parallel(system, num_trajectories=100, max_steps=10000, silent=False, processors=2):
    # This function only works in Python3
    import concurrent.futures as futures

    # executor = futures.ThreadPoolExecutor(max_workers=processors)
    executor = futures.ProcessPoolExecutor(max_workers=processors)

    futures_list = []
    for i in range(num_trajectories):
        futures_list.append(executor.submit(_run_trajectory, i, system, max_steps, silent))

    trajectories = []
    for f in futures.as_completed(futures_list):
        trajectories.append(f.result())

    return trajectories


def calculate_kmc_parallel_alternative(system, num_trajectories=100, max_steps=10000, silent=False, processors=2):

    from multiprocessing import cpu_count, Pool
    from functools import partial
    pool = Pool(processes=processors)
    trajectories = pool.map(partial(_run_trajectory, system=system, max_steps=max_steps, silent=silent), range(num_trajectories))
    return trajectories



import numpy as np


def kmc_algorithm(process_list):
    """
    :param rate_list: List with all the computed rates for all the neighbours for all the centers
    :param process_list: List of elements dict(center, process, new molecule).
    The indexes of each rate in rate_list have the same index that the associated process in
    process_list.
    Chooses a process using the list of rates and associates a time with this process using
    the BKL Kinetic Monte-Carlo algorithm. The algorithm uses 2 random number, one to choose the process and the other
    for the time. The usage of each random number is in an independent function
    :return:    plan: The chosen proces and the new molecule affected
                time: the duration of the process
    """

    rates_list = [proc.get_rate_constant() for proc in process_list]

    process_index = select_process(rates_list)
    chosen_process = process_list[process_index]

    time = time_advance(rates_list)

    return chosen_process, time


def select_process(rates_list):
    """
    :param rates_list: List with the constant rates
    :return: Chooses a position of the list chosen proportionally to its value.
    """
    r = np.sum(rates_list) * np.random.rand()
    # random number picked from the uniform distribution U(0, rates sum)

    sub_list = np.where(r > np.cumsum(rates_list))

    return len(sub_list[0])


def time_advance(rates_list):
    """
    :param rates_list: List with all the rates. Considering all the processes for all exciton
    :return: Process duration. Picks a random time from an exponential distribution
    """
    r = 1 - np.random.rand()  # interval [0, 1) -> (0, 1]
    return (-np.log(r)) / (np.sum(rates_list))



import numpy as np
from kimonet.utils import distance_vector_periodic
import inspect
from kimonet.utils.units import VAC_PERMITTIVITY
from kimonet.core.processes.transitions import Transition
from kimonet.utils.units import ATOMIC_TO_ANGS_EL
import kimonet.core.processes.forster as forster
from kimonet.utils import rotate_vector


coupling_data = {}


def generate_hash_2(function_name, d_transition, a_transition, d_orientation, a_orientation, r_vector, data_list):

    return hash((d_transition, a_transition,
                 d_orientation.tobytes(),
                 a_orientation.tobytes(),
                 function_name, tuple(data_list),
                 np.array(r_vector).tobytes()))


def generate_hash(function_name, initial, final, distance, data_list):

    return hash((initial, final, function_name,
                 function_name, tuple(data_list),
                 distance))


def forster_coupling(initial, final, ref_index=1, transitions=()):
    """
    Compute Forster coupling in eV

    :param initial: initial states
    :param final: final states
    :return: Forster coupling
    """

    d_transition = Transition(initial[0], final[1])
    a_transition = Transition(initial[1], final[0])

    d_orientation = initial[0].get_center().molecular_orientation()
    a_orientation = initial[1].get_center().molecular_orientation()

    #r_vector = initial[1].get_coordinates_absolute() - final[1].get_coordinates_absolute()
    r_vector = initial[0].get_coordinates_absolute() - final[0].get_coordinates_absolute()

    hash_string = generate_hash_2(inspect.currentframe().f_code.co_name,
                                  d_transition, a_transition,
                                  d_orientation, a_orientation,
                                  r_vector, [ref_index, tuple(transitions)])

    if hash_string in coupling_data:
        return coupling_data[hash_string]

    try:
        mu_a = transitions[transitions.index(a_transition)].tdm
        mu_d = transitions[transitions.index(d_transition)].tdm
    except ValueError:
        raise Exception('TDM for {} / {} not defined'.format(a_transition, d_transition))

    mu_d = rotate_vector(mu_d, d_orientation) * ATOMIC_TO_ANGS_EL
    mu_a = rotate_vector(mu_a, a_orientation) * ATOMIC_TO_ANGS_EL

    coupling_data[hash_string] = forster.dipole(mu_d, mu_a, r_vector, n=ref_index)

    return coupling_data[hash_string]


def forster_coupling_py(initial, final, ref_index=1, transitions=()):
    """
    Compute Forster coupling in eV

    :param initial: initial states
    :param final: final states
    :return: Forster coupling
    """

    d_transition = Transition(initial[0], final[1])
    a_transition = Transition(initial[1], final[0])

    d_orientation = initial[0].get_center().molecular_orientation()
    a_orientation = initial[1].get_center().molecular_orientation()

    # r_vector = initial[1].get_coordinates_absolute() - final[1].get_coordinates_absolute()
    r_vector = initial[0].get_coordinates_absolute() - final[0].get_coordinates_absolute()

    hash_string = generate_hash_2(inspect.currentframe().f_code.co_name,
                                  d_transition, a_transition,
                                  d_orientation, a_orientation,
                                  r_vector, [ref_index, tuple(transitions)])

    if hash_string in coupling_data:
        return coupling_data[hash_string]

    try:
        mu_a = transitions[transitions.index(a_transition)].tdm
        mu_d = transitions[transitions.index(d_transition)].tdm
    except ValueError:
        raise Exception('TDM for {} / {} not defined'.format(a_transition, d_transition))

    mu_d = rotate_vector(mu_d, d_orientation) * ATOMIC_TO_ANGS_EL
    mu_a = rotate_vector(mu_a, a_orientation) * ATOMIC_TO_ANGS_EL

    distance = np.linalg.norm(r_vector)

    k = orientation_factor(mu_d, mu_a, r_vector)              # orientation factor between molecules

    k_e = 1.0/(4.0*np.pi*VAC_PERMITTIVITY)

    coupling_data[hash_string] = k_e * k**2 * np.dot(mu_d, mu_a) / (ref_index**2 * distance**3)

    return coupling_data[hash_string]


def forster_coupling_extended(initial, final, ref_index=1, transitions=(), longitude=3, n_divisions=300):
    """
    Compute Forster coupling in eV

    :param initial: initial states
    :param final: final states
    :param longitude: extension length of the dipole
    :param n_divisions: number of subdivisions. To use with longitude. Increase until convergence.
    :return: Forster coupling
    """

    d_transition = Transition(initial[0], final[1])
    a_transition = Transition(initial[1], final[0])

    d_orientation = initial[0].get_center().molecular_orientation()
    a_orientation = initial[1].get_center().molecular_orientation()

    r_vector = initial[0].get_coordinates_absolute() - final[0].get_coordinates_absolute()

    hash_string = generate_hash_2(inspect.currentframe().f_code.co_name,
                                  d_transition, a_transition,
                                  d_orientation, a_orientation,
                                  r_vector, [ref_index, tuple(transitions), longitude, n_divisions])

    if hash_string in coupling_data:
        return coupling_data[hash_string]

    mu_a = transitions[transitions.index(a_transition)].tdm
    mu_d = transitions[transitions.index(d_transition)].tdm

    mu_d = rotate_vector(mu_d, d_orientation) * ATOMIC_TO_ANGS_EL
    mu_a = rotate_vector(mu_a, a_orientation) * ATOMIC_TO_ANGS_EL

    # mu_d = donor.get_transition_moment(to_state=_GS_)            # transition dipole moment (donor) e*angs
    # mu_a = acceptor.get_transition_moment(to_state=donor.state)  # transition dipole moment (acceptor) e*angs

    #r_vector = intermolecular_vector(donor, acceptor, supercell, cell_increment)  # position vector between donor and acceptor

    coupling_data[hash_string] = forster.dipole_extended(r_vector, mu_a, mu_d,
                                                         n=ref_index,
                                                         longitude=longitude,
                                                         n_divisions=n_divisions)

    return coupling_data[hash_string]


def forster_coupling_extended_py(initial, final, ref_index=1, transitions=(), longitude=3, n_divisions=300):
    """
    Compute Forster coupling in eV (pure python version)

    :param initial: initial states
    :param final: final states
    :param longitude: extension length of the dipole
    :param n_divisions: number of subdivisions. To use with longitude. Increase until convergence.
    :return: Forster coupling
    """

    d_transition = Transition(initial[0], final[1])
    a_transition = Transition(initial[1], final[0])

    d_orientation = initial[0].get_center().molecular_orientation()
    a_orientation = initial[1].get_center().molecular_orientation()

    r_vector = initial[0].get_coordinates_absolute() - final[0].get_coordinates_absolute()

    hash_string = generate_hash_2(inspect.currentframe().f_code.co_name,
                                  d_transition, a_transition,
                                  d_orientation, a_orientation,
                                  r_vector, [ref_index, tuple(transitions), longitude, n_divisions])

    if hash_string in coupling_data:
        return coupling_data[hash_string]

    mu_a = transitions[transitions.index(a_transition)].tdm
    mu_d = transitions[transitions.index(d_transition)].tdm

    mu_d = rotate_vector(mu_d, d_orientation) * ATOMIC_TO_ANGS_EL
    mu_a = rotate_vector(mu_a, a_orientation) * ATOMIC_TO_ANGS_EL

    # mu_d = donor.get_transition_moment(to_state=_GS_)            # transition dipole moment (donor) e*angs
    # mu_a = acceptor.get_transition_moment(to_state=donor.state)  # transition dipole moment (acceptor) e*angs

    #r_vector = intermolecular_vector(donor, acceptor, supercell, cell_increment)  # position vector between donor and acceptor

    mu_ai = mu_a / n_divisions
    mu_di = mu_d / n_divisions

    k_e = 1.0 / (4.0 * np.pi * VAC_PERMITTIVITY)

    forster_coupling = 0
    for x in np.linspace(-0.5 + 0.5/n_divisions, 0.5 - 0.5/n_divisions, n_divisions):
        for y in np.linspace(-0.5 + 0.5/n_divisions, 0.5 - 0.5/ n_divisions, n_divisions):

            #print(x, y)
            dr_a = mu_a / np.linalg.norm(mu_a) * longitude * x
            dr_d = mu_d / np.linalg.norm(mu_d) * longitude * y
            r_vector_i = r_vector + dr_a + dr_d

            distance = np.linalg.norm(r_vector_i)

            k = orientation_factor(mu_ai, mu_di, r_vector_i)              # orientation factor between molecules

            forster_coupling += k_e * k**2 * np.dot(mu_ai, mu_di) / (ref_index**2 * distance**3)

    coupling_data[hash_string] = forster_coupling                            # memory update for new couplings

    return forster_coupling


def intermolecular_vector(molecule_1, molecule_2, supercell, cell_incr):
    """
    :param molecule_1: donor
    :param molecule_2: acceptor
    :return: the distance between the donor and the acceptor
    """

    position_d = molecule_1.get_coordinates()
    position_a = molecule_2.get_coordinates()
    r_vector = position_a - position_d
    r = distance_vector_periodic(r_vector, supercell, cell_incr)
    return r


def orientation_factor(u_d, u_a, r):
    """
    :param u_d: dipole transition moment of the donor
    :param u_a: dipole transition moment of the acceptor
    :param r:  intermolecular_distance
    :type u_d: np.ndarray
    :type u_a: np.ndarray
    :type r: float

    :return: the orientational factor between both molecules
    :rtype: float
    """
    nd = unit_vector(u_d)
    na = unit_vector(u_a)
    e = unit_vector(r)
    return np.dot(nd, na) - 3*np.dot(e, nd)*np.dot(e, na)


def unit_vector(vector):
    """
    :param vector:
    :return: computes a unity vector in the direction of vector
    """
    return vector / np.linalg.norm(vector)


def dexter_coupling(initial, final, k_factor=1):
    """
    Compute Dexter coupling in eV

    :param initial: initial states
    :param final: final states
    :param vdw_radius: dictionary with Van der Waals radius for each state
    :param k_factor: Dexter parameter K
    :return: Dexter coupling
    """

    function_name = inspect.currentframe().f_code.co_name

    r_vector = initial[0].get_coordinates_absolute() - final[0].get_coordinates_absolute()
    distance = np.linalg.norm(r_vector)

    # donor <-> acceptor interaction symmetry
    hash_string = generate_hash(function_name, initial, final, distance, [k_factor])

    if hash_string in coupling_data:
        return coupling_data[hash_string]

    vdw_radius_sum = initial[0].vdw_radius + final[0].vdw_radius

    # print('dexter: ', 2/vdw_radius_sum)
    dexter_coupling = k_factor * np.exp(-2 * distance / vdw_radius_sum)

    coupling_data[hash_string] = dexter_coupling                            # memory update for new couplings

    return dexter_coupling




from copy import deepcopy
from kimonet.system.state import ground_state as _GS_
from kimonet.utils.combinations import combinations_group
import numpy as np


def get_processes(state, system):
    """
    :param state: excited state for which find possible processes
    :param system: Instance of System class
    Computes the transfer and decay rates and builds two dictionaries:
            One with the decay process as key and its rate as argument
            One with the transferred molecule index as key and {'process': rate} as argument

    :return:    process_list: List of processes in named tuple format
                rate_list: List with the respective rates
    """

    transfer_processes = get_transfer_rates(state, system)
    decay_processes = get_decay_rates(state, system)

    return decay_processes + transfer_processes


def get_transfer_rates(state, system):
    """
    :param center: Index of the studies excited molecule
    :param system: Dictionary with the list of molecules and additional physical information
    :return: Two lists, one with the transfer rates and the other with the transfer processes.
    """

    neighbors, cell_increments = system.get_state_neighbors(state)

    if False:
        print('**** SYSTEM STATE INTER ****')
        for i, mol in enumerate(system.molecules):
            print(i, mol.state.label, mol.state, mol.state.cell_state, mol.name, mol)
        print('----')
        for state in system.get_states():
            print(state.label, state, state.supercell)
        print('****************')

    transfer_steps = []
    for acceptor_state, cell_incr in zip(neighbors, cell_increments):
        allowed_processes = get_allowed_processes(state, acceptor_state, system.transfer_scheme, cell_incr)
        for process in allowed_processes:
            transfer_steps.append(process)

    return transfer_steps


def get_decay_rates(state, system):
    """
    :param center: index of the excited molecule
    :param system: Dictionary with all the information of the system
    :return: A dictionary with the possible decay rates
    For computing them the method get_decay_rates of class molecule is call.
    """

    decay_complete = system.decay_scheme

    decay_steps = []
    for process in decay_complete:

        if process.initial[0].label == state.label:

            elements_list = state.get_molecules()
            # group_list = [s.size for s in process.final_test]

            configurations = combinations_group(elements_list, process.final_test, supercell=process.supercell)

            for configuration in configurations:

                new_process = deepcopy(process)
                new_process.initial = (state,)

                for molecules, final in zip(configuration, new_process.final_test):
                    final.remove_molecules()
                    final.cell_state = np.zeros_like(molecules[0].cell_state) # set zero to final state cell_states
                    for mol in molecules:
                        new_process.cell_states[mol] = mol.cell_state  # keep molecules with same cell_states
                        final.add_molecule(mol)

                new_process.reset_cell_states()

                # print(new_process.initial[0].label, '--', new_process.final[0].label)
                decay_steps.append(new_process)

    return decay_steps


def get_allowed_processes(donor_state, acceptor_state, transfer_scheme, cell_incr):
    """
    Get the allowed processes for a given donor and acceptor

    :param donor_state: State instance containing the donor state
    :param acceptor_state: State instance containing the acceptor state
    :param cell_incr: List with the cell_state increments for the given acceptor (diff between acceptor and donor cell states)
    :return: Dictionary with the allowed process functions
    """

    allowed_processes = []
    for process in transfer_scheme:

        # handle self interaction
        if donor_state == acceptor_state:
            process = process.get_self_interaction_process()
            if process is None:
                continue

        if (process.initial[0].label, process.initial[1].label) == (donor_state.label, acceptor_state.label):

            def is_same_p_configuration(p_configuration_1, p_configuration_2):
                if len(p_configuration_1) != len(p_configuration_2):
                    return False

                sum1 = np.multiply(*[hash(state) for state in p_configuration_1])
                sum2 = np.multiply(*[hash(state) for state in p_configuration_2])

                return sum1 == sum2

            include_self = not is_same_p_configuration(process.initial, process.final_test)

            # Getting all possible configurations for the final states
            elements_list = [state.get_molecules() for state in (donor_state, acceptor_state)]
            elements_list = [item for sublist in elements_list for item in sublist]
            # group_list = [state.size for state in process.final_test]

            configurations = combinations_group(elements_list, process.final_test, supercell=process.supercell, include_self=include_self)

            for configuration in configurations:

                new_process = deepcopy(process)
                new_process.initial = (donor_state, acceptor_state)
                new_process.set_cell_increment(cell_incr)

                # Binding final states to initial states if equal
                #for final in new_process.final_test:
                #    final.cell_state = np.zeros_like(donor_state.cell_state)
                #    for initial in new_process.initial:
                #        if initial.label == final.label and initial.label != _GS_.label:
                #            print(final, initial)
                #            final.cell_state = initial.cell_state

                for final in new_process.final_test:
                    final.cell_state = np.zeros_like(donor_state.cell_state)
                for initial, final_list in new_process.get_transport_connections().items():
                    for final in final_list:
                        final.cell_state = initial.cell_state

                # Binding molecules to states
                for molecules, final in zip(configuration, new_process.final_test):
                    final.remove_molecules()
                    for mol in molecules:
                        final.add_molecule(mol)

                # redefine cell states for molecules in the process
                for mol in new_process.get_molecules():
                    new_process.cell_states[mol] = np.array(mol.cell_state)
                    if mol in new_process.initial[0].get_molecules():
                        new_process.cell_states[mol] += np.array(cell_incr)
                    elif mol in new_process.initial[1].get_molecules():
                        new_process.cell_states[mol] -= np.array(cell_incr)
                    else:
                        raise Exception('molecules from initial and final do not match')

                # redefine state cell_state from molecules cell_state
                for initial, final in zip(new_process.initial, new_process.final_test):
                    if final.label != _GS_.label:
                        cell_diff = np.array(np.average([new_process.cell_states[mol] for mol in final.get_molecules()], axis=0),
                                             dtype=int)

                        for mol in final.get_molecules():
                            new_process.cell_states[mol] -= cell_diff

                        final.cell_state += cell_diff

                allowed_processes.append(new_process)

    return allowed_processes



import numpy as np
from numpy import pi
from kimonet.system.molecule import Molecule
from kimonet.utils.units import BOLTZMANN_CONSTANT


def theoretical_diffusion_values(system_information):
    """
    :param system_information: dictionary with the system information (parameter to parameter)
    :return: dictionary with the theoretical lifetime, diffusion constant and length. (In the cases that they can be
    defined).
    """

    # system information
    k = 2                         # orientational factor. Always 2 in the systems where this model is possible

    r = system_information['lattice']['lattice_parameter'][0]       # equall in all directions (nm)

    T = system_information['conditions']['temperature']             # temperature (K)
    n = system_information['conditions']['refractive_index']        # refractive index

    d = len(system_information['lattice']['dimensions'])                       # dimensionality of the system

    ############################################################################################################

    # molecule information.     Again a generic object of molecule is initialized

    excited_state = list(system_information['excitons'].keys())[0]
    # in a diffusion study there will be only on exciton, so only one key

    reorganization = system_information['reorganization_energies'][excited_state]
    # reorganization energy of the excited electronic state. eV

    transition_moment_vector = system_information['transition_moment']          # transition dipole moment (a.u)
    transition_moment = np.linalg.norm(transition_moment_vector)                # modulus

    # initialization of a generic instance of class molecule
    generic_molecule = Molecule(state_energies=system_information['state_energies'],
                                state='s1', reorganization_energies=system_information['reorganization_energies'],
                                transition_moment=transition_moment_vector)

    ############################################################################################################

    # The life time of the exciton is taken as the inverse of the sum of all decay rates
    # A theoretical value for the life time will be always computed.
    decay_rates = generic_molecule.decay_rates()
    decay_sum = 0
    for decay in decay_rates:
        decay_sum =+ decay_rates[decay]
    life_time = 1 / decay_sum

    ############################################################################################################

    # cases in which there is a theoretical model for the diffusion
    possible_cases = [['1d_ordered', 'parallel'], ['2d_ordered', 'parallel'], ['3d_ordered', 'parallel'],
                      ['1d_ordered', 'antiparallel'], ['2d_ordered', 'antiparallel'], ['3d_ordered', 'antiparallel']]

    # the theoretical values for D and LD will always be computed for the possible cases:
    if [system_information['type'], system_information['orientation']] in possible_cases:

        factor_1 = 2 * pi * (transition_moment**2 * k**2 / ((r/0.053)**3 * n**2))**2    # in atomic units
        factor_2 = np.sqrt(1 / (4 * pi * reorganization * BOLTZMANN_CONSTANT * T)) * \
                   np.exp(- reorganization / (4 * BOLTZMANN_CONSTANT * T))                  # in eV

        rate = factor_1 * factor_2

        D = rate * r**2                                             # diffusion constant (nm² ns⁻¹)
        Ld = np.sqrt(2 * d * rate * r**2 * life_time)               # diffusion length (nm)

    else:                                                           # Ld and D taken as None if there is no
        D = None                                                    # theoretical model
        Ld = None

    # return the 3 parameters in a dictionary
    return {'lifetime': life_time, 'diffusion_constant':  D,
            'diffusion_length': Ld}



import numpy as np
import warnings
from kimonet.utils.units import BOLTZMANN_CONSTANT, HBAR_PLANCK
import scipy.integrate as integrate
import scipy.interpolate as interpolate
from kimonet.core.processes.transitions import Transition


class MarcusModel:

    def __init__(self,
                 transitions=None,
                 temperature=300,  # Kelvin
                 drift_term=0  # aN -> eV / angs [to be implemented]
                 ):


        self.temperature = temperature
        self._transitions = transitions


    def __hash__(self):
        #return hash((str(self.reorganization_energies)))
        return hash((str(self._transitions)))

    def get_vib_spectrum(self, target_state, origin_state):

        # elec_trans_ene = self.state_energies[transition[1]] - self.state_energies[transition[0]]
        elec_trans_ene = target_state.energy - origin_state.energy
        transition = Transition(target_state, origin_state, symmetric=False)

        temp = self.temperature  # temperature (K)

        if transition in self._transitions:
            reorg_ene = self._transitions[self._transitions.index(transition)].reorganization_energy
        #if transition in self.reorganization_energies:
        #    reorg_ene = np.sum(self.reorganization_energies[transition])
        else:
            raise Exception('Reorganization energy for transition {} not defined'.format(transition))

        sign = np.sign(elec_trans_ene)

        # print('T', temp, molecule.reorganization_energies[transition], elec_trans_ene, -sign)

        def vib_spectrum(e):
            return 1.0 / (np.sqrt(4.0 * np.pi * BOLTZMANN_CONSTANT * temp * reorg_ene)) * \
                   np.exp(-(elec_trans_ene - e * sign + reorg_ene) ** 2 / (4 * BOLTZMANN_CONSTANT * temp * reorg_ene))

        return vib_spectrum


class LevichJortnerModel:

    def __init__(self,
                 frequencies=None,
                 reorganization_energies=None,  # eV
                 external_reorganization_energies=None,  # eV
                 temperature=300,
                 # state_energies=None,
                 ):

        self.frequencies = frequencies
        self.reorganization_energies = reorganization_energies
        self.external_reorganization_energies = external_reorganization_energies
        self.state_energies = None
        self.temperature = temperature

        # symmetrize external reorganization energies
        """
        if self.external_reorganization_energies is not None:
            for key in list(external_reorganization_energies):
                external_reorganization_energies[key[::-1]] = external_reorganization_energies[key]
        """

    def __hash__(self):
        return hash((str(self.state_energies),
                     str(self.frequencies),
                     str(self.external_reorganization_energies),
                     str(self.reorganization_energies)))

    def set_state_energies(self, state_energies):
        self.state_energies = state_energies

    def get_vib_spectrum(self, target_state, origin_state):

        elec_trans_ene = target_state.energy - origin_state.energy
        transition = Transition(target_state, origin_state, symmetric=False)

        #elec_trans_ene = self.state_energies[transition[1]] - self.state_energies[transition[0]]

        temp = self.temperature  # temperature (K)
        ext_reorg_ene = self.external_reorganization_energies[transition]
        reorg_ene = np.array(self.reorganization_energies[transition])

        sign = np.sign(elec_trans_ene)

        angl_freqs = np.array(self.frequencies[transition]) * 29.9792558 * 2*np.pi  # cm-1 -> ns-1 , angular frequency

        freq_classic_limit = BOLTZMANN_CONSTANT * temp / HBAR_PLANCK  # ns^-1,  angular frequency

        indices_cl = np.where(angl_freqs < freq_classic_limit)[0]
        indices_qm = np.where(angl_freqs >= freq_classic_limit)[0]

        l_cl = np.sum(reorg_ene[indices_cl]) + ext_reorg_ene

        l_qm = reorg_ene[indices_qm]

        ang_freq_qm = angl_freqs[indices_qm]

        s = np.true_divide(l_qm, ang_freq_qm) / HBAR_PLANCK

        s_eff = np.sum(s)
        w_eff = np.sum(np.multiply(s, ang_freq_qm))/s_eff  # angular frequency

        # print('l_qm', np.sum(l_qm))
        # print('s_eff', s_eff)
        # print('w_eff', w_eff)

        def vib_spectrum(e):
            e = np.array(e, dtype=float)
            fcwd_term = np.zeros_like(e)
            for m in range(10):
                fcwd_term += s_eff**m / np.math.factorial(m) * np.exp(-s_eff) * np.exp(
                    -(elec_trans_ene - e * sign + l_cl + m * HBAR_PLANCK * w_eff)**2 / (4 * BOLTZMANN_CONSTANT * temp * l_cl))

            return 1.0 / (np.sqrt(4 * np.pi * BOLTZMANN_CONSTANT * temp * l_cl)) * fcwd_term

        return vib_spectrum


class EmpiricalModel:

    def __init__(self,
                 empirical_function=None,  # eV
                 ):

        self.empirical_function = empirical_function
        self.state_energies = None

    def __hash__(self):
        return hash((str(self.state_energies),
                     str(self.empirical_function)))

    def set_state_energies(self, state_energies):
        self.state_energies = state_energies

    def get_vib_spectrum(self, target_state, origin_state):
        transition = Transition(target_state, origin_state, symmetric=False)

        # Temperature is not actually used. This is to keep common interface
        return self.empirical_function[transition]


class GaussianModel:

    def __init__(self,
                 deviations=None,
                 reorganization_energies=None,  # eV
                 # state_energies=None,
                 ):

        self.reorganization_energies = reorganization_energies
        self.state_energies = None
        self.deviations = deviations

        # symmetrize external reorganization energies
        """
        if self.external_reorganization_energies is not None:
            for key in list(external_reorganization_energies):
                external_reorganization_energies[key[::-1]] = external_reorganization_energies[key]
        """

    def __hash__(self):
        return hash((str(self.state_energies),
                     str(self.deviations),
                     str(self.reorganization_energies)))

    def get_vib_spectrum(self, target_state, origin_state):

        """
        :param donor: energy diference between states
        :param acceptor: deviation in energy units
        :return: Franck-Condon-weighted density of states in gaussian aproximation
        """

        elec_trans_ene = target_state.energy - origin_state.energy
        transition = Transition(target_state, origin_state, symmetric=False)

        #elec_trans_ene = self.state_energies[transition[1]] - self.state_energies[transition[0]]
        reorg_ene = np.sum(self.reorganization_energies[transition])

        deviation = self.deviations[transition]     # atomic units

        sign = np.sign(elec_trans_ene)
        def vib_spectrum(e):
            return np.exp(-(elec_trans_ene - e * sign + reorg_ene) ** 22 / (2 * deviation) ** 2) / (2 * np.sqrt(np.pi) * deviation)

        return vib_spectrum


class NoVibration:

    def __init__(self,
                 ):

        self.state_energies = None

    def __hash__(self):
        return hash((str(self.state_energies)))

    def set_state_energies(self, state_energies):
        self.state_energies = state_energies

    def get_vib_spectrum(self, target_state, origin_state):

        elec_trans_ene = target_state.energy - origin_state.energy

        return elec_trans_ene

        #transition = Transition(target_state, origin_state, symmetric=False)

        # elec_trans_ene = self.state_energies[transition[1]] - self.state_energies[transition[0]]

        #def vib_spectrum(e):
        #    if elec_trans_ene == e:
        #        return 1
        #    else:
        #        return 0
        #
        #return vib_spectrum


# Analysis functions
def get_normalized_spectrum(x, y, in_nm=False, interpolation='quadratic'):

    warnings.filterwarnings("ignore", category=integrate.IntegrationWarning)

    _EV_TO_NM_ = 1239.841

    if in_nm:
        x = _EV_TO_NM_ / np.array(x)

    # from scipy.integrate import quad
    # def integrand(x, donor, acceptor):
    #    return f_d(x) * f_a(x)
    # fcwd = quad(integrand, 1.e-8, np.inf, args=(0, 0),  epsabs=1e-20)[0]

    # Normalize spectra
    normalization_constant = abs(integrate.simps(y, x))
    print('normalize', normalization_constant)

    y /= normalization_constant

    f_x = interpolate.interp1d(x, y, fill_value=0, bounds_error=False, kind=interpolation)

    return f_x


def print_data(f_donor, f_acceptor):

    def overlap(x):
        return f_donor(x) * f_acceptor(x)

    x_new = np.linspace(1.5, 4, 1000)

    fcwd_test = integrate.simps(f_donor(x_new) * f_acceptor(x_new), x_new)

    test_donor = integrate.quad(f_donor, 0, np.inf,  epsabs=1e-20)[0]
    test_acceptor = integrate.quad(f_acceptor, 0, np.inf,  epsabs=1e-20)[0]
    test_overlap = integrate.quad(overlap, 0, np.inf,  epsabs=1e-20)[0]

    print('Integral, Donor: {} Acceptor: {} Overlap: {}'.format(test_donor, test_acceptor, test_overlap))
    print('FCWD : {}'.format(fcwd_test))

    center_em = np.average(x_new, weights=f_donor(x_new))
    center_abs = np.average(x_new, weights=f_acceptor(x_new))

    print('Center abs: {} eV'.format(center_abs))
    print('Center em: {} eV'.format(center_em))

    print('Reorganization energy: {} eV'.format(np.abs(center_abs - center_em)))



import numpy as np


def rot_x(angle):
    """
    rotate with respect x axis (right hand criteria)
    :param angle:
    :return:
    """
    return np.array([[1,             0,              0],
                     [0, np.cos(angle), -np.sin(angle)],
                     [0, np.sin(angle),  np.cos(angle)]])


def rot_y(angle):
    """
    rotate with respect y axis (right hand criteria)
    :param angle:
    :return:
    """
    return np.array([[np.cos(angle),  0, np.sin(angle)],
                     [0,              1,             0],
                     [-np.sin(angle), 0, np.cos(angle)]])


def rot_z(angle):
    """
    rotate with respect z axis (right hand criteria)
    :param angle:
    :return:
    """
    return np.array([[np.cos(angle), -np.sin(angle), 0],
                     [np.sin(angle),  np.cos(angle), 0],
                     [0,                          0, 1]])


def rotate_vector_3d(vector, ang_x, ang_y, ang_z):
    return np.dot(rot_z(ang_z), np.dot(rot_y(ang_y), np.dot(rot_x(ang_x), vector)))


def rotate_vector(vector, orientation):
    """
    Rotate vector of dimensions (1-3)

    :param vector: vector to rotate
    :param orientation: angles (in radians) to rotate respect to axis x, y, z
    :return: rotated vector
    """

    norm = np.linalg.norm(vector)
    n_dim = len(vector)
    ang_x, ang_y, ang_z = orientation
    rotated_vector = np.dot(rot_z(ang_z)[:n_dim, :n_dim],
                        np.dot(rot_y(ang_y)[:n_dim, :n_dim],
                               np.dot(rot_x(ang_x)[:n_dim, :n_dim],
                                      vector)))

    # Rotated vector is defined to have same norm as original whatever orientation
    return rotated_vector/np.linalg.norm(rotated_vector) * norm



from kimonet.core.kmc import kmc_algorithm
from kimonet.core.processes import get_processes
from kimonet.core.processes.types import GoldenRule, DirectRate, DecayRate
import numpy as np


def do_simulation_step(system):
    """
    :param system: system object
    :return: the chosen process and time
    """

    if False:
        print('**** SYSTEM STATE INITIAL ****')
        system.print_status()
        print('****************')

    process_collector = []                          # list with the respective processes (for all centers)
    for state in system.get_states():
        process_collector += get_processes(state, system)

    # If no process available system cannot evolve and simulation is finished
    if len(process_collector) == 0:
        system.is_finished = True
        return None, 0
    chosen_process, time = kmc_algorithm(process_collector)

    system.update(chosen_process)

    if False:
        print('**** SYSTEM STATE FINAL ****')
        system.print_status()
        print('****************')

    return chosen_process, time


def system_test_info(system):
    for state in system.get_states():
        center = system.get_molecule_index(state.get_center())
        print('*' * 80 + '\n CENTER {}\n'.format(center) + '*' * 80)

        process_list = get_processes(state, system)
        total_r = 0
        for proc in process_list:

            i_donor = system.get_molecule_index(proc.initial[0].get_center())
            try:
                i_acceptor = system.get_molecule_index(proc.initial[1].get_center())
            except Exception:
                i_acceptor = i_donor

            print('Description:', proc.description)
            print('Donor: {} / Acceptor: {}'.format(i_donor, i_acceptor))

#            position_d = proc.initial[0].get_center().get_coordinates()
            r = proc.get_rate_constant()

            # if isinstance(proc, (GoldenRule, DirectRate)):
            if len(proc.initial) == 2:
                cell_increment = proc.initial_absolute[1].cell_state - proc.initial_absolute[0].cell_state
                distance = np.linalg.norm([proc.initial_absolute[1].get_coordinates_absolute() - proc.initial_absolute[0].get_coordinates_absolute()])
                print('Distance: {:.4} angs'.format(distance))
                print('Cell_increment: {} '.format(cell_increment))

            if isinstance(proc, GoldenRule):

                spectral_overlap = proc.get_fcwd()

                e_coupling = proc.get_electronic_coupling()

                print('Electronic coupling: {:.4} eV'.format(e_coupling))
                print('Spectral overlap:    {:.4} eV-1'.format(spectral_overlap))
                # anal_data.append([distance, r])

            print('Rate constant: {:.4} ns-1'.format(r))

            print('-' * 80)
            total_r += r

        print('Total rate sum: {:.4}'.format(float(total_r)))

        # import matplotlib.pyplot as plt
        # plt.scatter(np.array(anal_data).T[0], np.array(anal_data).T[1])
        # plt.show()




import numpy as np
import matplotlib.pyplot as plt


def normalize_cell(supercell):
    normalize = []
    for r in np.array(supercell):
        normalize.append(r/np.linalg.norm(r))
    return np.array(normalize)


class TrajectoryAnalysis:

    def __init__(self, trajectories):
        self.trajectories = trajectories
        self.n_dim = trajectories[0].get_dimension()
        self.n_traj = len(trajectories)

        self.states = set()
        for traj in trajectories:
            self.states |= traj.get_states()
        self.states = self.states
        self._points_ratio = {}
        self._segment_ratio = {}

    def __str__(self):

        txt_data = '\nTrajectory Analysis\n'
        txt_data += '------------------------------\n'
        txt_data += 'Number of trajectories: {}\n'.format(self.n_traj)
        txt_data += 'Dimension: {}\n'.format(self.n_dim)
        txt_data += 'Number of nodes: {}\n'.format(self.get_number_of_nodes())
        txt_data += 'States: {}\n'.format(self.states)

        return txt_data

    def get_number_of_nodes(self):
        return len([traj.get_number_of_nodes() for traj in self.trajectories])

    def get_states(self):
        return self.states

    def get_lifetime_ratio(self, state):
        return np.average([traj.get_time_ratio(state) for traj in self.trajectories])

    def get_segment_ration(self, state=None):
        if state not in self._segment_ratio:
            sum_t = np.nansum([traj.get_n_segments(state) for traj in self.trajectories])
            if sum_t != 0:
                self._segment_ratio[state] = [traj.get_n_segments(state)/float(sum_t) for traj in self.trajectories]
            else:
                self._segment_ratio[state] = None

        return self._segment_ratio[state]


    def diffusion_coeff_tensor(self, state, unit_cell=None):
        """
        calculate the average diffusion tensor defined as:

        DiffTensor = 1/2 * <DiffLen^2> / <time>

        :param state: electronic state to analyze
        :return:
        """

        total_traj = np.sum([traj.get_n_subtrajectories(state) for traj in self.trajectories])
        if total_traj == 0:
            return None

        tensor_list = [traj.get_diffusion_tensor(state)*traj.get_n_subtrajectories(state) for traj in
                       self.trajectories if traj.get_diffusion_tensor(state) is not None]

        tensor = np.sum(tensor_list, axis=0)/total_traj

        if unit_cell is not None:
            trans_mat = normalize_cell(unit_cell)
            mat_inv = np.linalg.inv(trans_mat)
            # tensor = np.dot(mat_inv.T, tensor)
            tensor = np.dot(np.dot(mat_inv.T, tensor), mat_inv)

        return tensor

    def diffusion_length_square_tensor(self, state, unit_cell=None):
        """
        calculate the average diffusion length tensor defined as:

        DiffLenTen = 2 * DiffTensor * lifetime

        :param state: electronic state to analyze
        :return:
        """

        total_traj = np.sum([traj.get_n_subtrajectories(state) for traj in self.trajectories])
        if total_traj == 0:
            return None

        tensor_list = [traj.get_diffusion_length_square_tensor(state)*traj.get_n_subtrajectories(state)
                       for traj in self.trajectories if traj.get_diffusion_length_square_tensor(state) is not None]

        tensor = np.sum(tensor_list, axis=0)/total_traj

        if unit_cell is not None:
            trans_mat = normalize_cell(unit_cell)
            mat_inv = np.linalg.inv(trans_mat)
            # tensor = np.dot(mat_inv.T, tensor)
            tensor = np.dot(np.dot(mat_inv.T, tensor), mat_inv)

        return tensor

    def diffusion_coefficient_old(self, state=None):
        """
        Return the average diffusion coefficient defined as:

        DiffCoeff = 1/(2*z) * <DiffLen^2>/<time>

        :return:
        """

        if state is None:
            sum_diff = 0
            sum_n_subtraj = 0
            for istate in self.get_states():
                n_subtraj = len(self.get_segment_ration(istate))

                diffusion_list = [traj.get_diffusion(istate)*s for traj, s in zip(self.trajectories, self.get_segment_ration(istate))
                                  if istate in traj.get_states()]
                if not np.isnan(diffusion_list).all():
                    sum_diff += np.nansum(diffusion_list) * n_subtraj
                    # sum_diff += np.nanmean(diffusion_list) * self.get_lifetime_ratio(s)
                    sum_n_subtraj += n_subtraj
            return sum_diff/sum_n_subtraj

        if self.get_segment_ration(state) is None:
            return None

        return np.nansum([traj.get_diffusion(state)*s for traj, s in zip(self.trajectories, self.get_segment_ration(state))])
        # return np.nanmean([traj.get_diffusion(state) for traj in self.trajectories])

    def diffusion_coefficient(self, state=None):
        """
        Return the average diffusion coefficient defined as:

        DiffCoeff = 1/(2*z) * <DiffLen^2>/<time>

        :return:
        """

        total_traj = np.sum([traj.get_n_subtrajectories(state) for traj in self.trajectories])
        if total_traj == 0:
            return None

        return np.sum([traj.get_diffusion(state)*traj.get_n_subtrajectories(state) for traj in self.trajectories if
                       traj.get_diffusion(state) is not None])/total_traj

    def lifetime_old(self, state=None):

        if state is None:
            sum_diff = 0
            sum_n_subtraj = 0
            for istate in self.get_states():
                n_subtraj = len(self.get_segment_ration(istate))
                lifetime_list = [traj.get_lifetime(istate)*s for traj, s in zip(self.trajectories, self.get_segment_ration(istate))
                                 if istate in traj.get_states()]
                # diffusion_list = [traj.get_lifetime(s) for traj in self.trajectories]
                if not np.isnan(lifetime_list).all():
                    # sum_diff += np.nanmean(diffusion_list) * self.get_lifetime_ratio(s)
                    sum_diff += np.nansum(lifetime_list) * n_subtraj
                    sum_n_subtraj += n_subtraj

            return sum_diff/sum_n_subtraj

        if self.get_segment_ration(state) is None:
            return None

        return np.nansum([traj.get_lifetime(state)*s for traj, s in zip(self.trajectories, self.get_segment_ration(state))])
        # return np.average([traj.get_lifetime(state) for traj in self.trajectories])

    def lifetime(self, state=None):

        total_traj = np.sum([traj.get_n_subtrajectories(state) for traj in self.trajectories])
        if total_traj == 0:
            return None

        return np.sum([traj.get_lifetime(state)*traj.get_n_subtrajectories(state) for traj in self.trajectories if
                       traj.get_lifetime(state) is not None])/total_traj

    def diffusion_length(self, state=None):
        """
        Return the average diffusion coefficient defined as:

        DiffLen = SQRT(2 * z * DiffCoeff * LifeTime)

        :return:
        """

        total_traj = np.sum([traj.get_n_subtrajectories(state) for traj in self.trajectories])
        if total_traj == 0:
            return None

        length2 = np.sum([traj.get_diffusion_length_square(state)*traj.get_n_subtrajectories(state) for traj in self.trajectories
                          if traj.get_diffusion_length_square(state) is not None])/total_traj

        return np.sqrt(length2)

    def plot_2d(self, state=None):
        plt = None
        for traj in self.trajectories:
            plt = traj.plot_2d(state, show_warnings=False)
        return plt

    def plot_distances(self, state=None):
        plt = None
        for traj in self.trajectories:
            plt = traj.plot_distances(state)
        return plt

    def plot_exciton_density(self, state=None):

        time_max = np.max([traj.get_simulation_times()[-1] for traj in self.trajectories]) * 1.1
        t_range = np.linspace(0, time_max, 100)

        ne_interp = []
        for traj in self.trajectories:
            ne = traj.get_number_of_excitons(state)
            t = traj.get_simulation_times()
            ne_interp.append(np.interp(t_range, t, ne, right=0))

        plt.title('Averaged exciton number ({})'.format('' if state is None else state))
        plt.ylim(bottom=0, top=np.max(ne_interp))
        plt.xlim(left=0, right=time_max)
        plt.xlabel('time (ns)')
        plt.ylabel('# of excitons in supercell')
        plt.plot(t_range, np.average(ne_interp, axis=0), label='Total' if state is None else state)
        plt.legend()
        return plt

    def plot_histogram(self, state=None, normalized=False, bins=None):

        distances = []
        for traj in self.trajectories:
            d, _ = traj.get_max_distances_vs_times(state)
            distances += list(d)

        plt.title('Distances histogram  ({})'.format('' if state is None else state))
        plt.xlabel('Distance (Angs)')
        if normalized:
            plt.ylabel('Probability density (Angs^-1)')
        else:
            plt.ylabel('# of occurrences')
        try:
            plt.hist(distances, density=normalized, bins=bins)
        except AttributeError:
            plt.hist(distances, normed=normalized, bins=bins)
        return plt


def _get_diffusion_helper(trajectory, state):
    return trajectory.get_diffusion(state)


class TrajectoryAnalysisParallel(TrajectoryAnalysis):
    def __init__(self, trajectories, processors=2):
        super().__init__(trajectories)
        import concurrent.futures as futures
        self._executor = futures.ProcessPoolExecutor(max_workers=processors)
        # self._executor = futures.ThreadPoolExecutor(max_workers=processors)

    def diffusion_coefficient(self, state=None):
        """
        Return the average diffusion coefficient defined as:
        1/(2*z) * <DiffLen^2>/<time>
        *Parallel version*

        :param state: state label
        :return:
        """
        import concurrent.futures as futures

        if state is None:
            sum_diff = 0
            for s in self.get_states():
                futures_list = []
                for trajectory in self.trajectories:
                    futures_list.append(self._executor.submit(_get_diffusion_helper, trajectory, s))

                diffusion_list = []
                for f in futures.as_completed(futures_list):
                    diffusion_list.append(f.result())

                if not np.isnan(diffusion_list).all():
                    sum_diff += np.nanmean(diffusion_list) * self.get_lifetime_ratio(s)
            return sum_diff

        futures_list = []
        for trajectory in self.trajectories:
            futures_list.append(self._executor.submit(_get_diffusion_helper, trajectory, state))

        diffusion_list = []
        for f in futures.as_completed(futures_list):
            diffusion_list.append(f.result())

        return np.nanmean(diffusion_list)



# from kimonet.core.processes import Transfer
import networkx as nx
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.cm as cm
from scipy import stats
import warnings
import os
from copy import deepcopy
from kimonet.system.state import ground_state as _GS_


def count_keys_dict(dictionary, key):
    if key in dictionary:
        dictionary[key] += 1
    else:
        dictionary[key] = 1


class ArrayHandler():
    def __init__(self, arraystructure, data):
        self.arraystructure = arraystructure

        if data in self.arraystructure.dtype.fields:
            self.data = data
        else:
            raise Exception('Data not in array')

        # determine data position
        for i, element in enumerate(self.arraystructure.dtype.fields):
            if self.data == element:
                self.index = i

        # set initial data len
        self.data_len = len(arraystructure)

    def __str__(self):
        return str(self.arraystructure[self.data][:self.data_len])

    def __getitem__(self, item):
        return self.arraystructure[self.data][:self.data_len][item]

    def __len__(self):
        return self.data_len

    def append(self, value):

        if self.data_len >= len(self.arraystructure):
            self.arraystructure.resize((len(self.arraystructure) + 1,), refcheck=False)
            new_element = [None] * len(self.arraystructure.dtype.fields)
        else:
            new_element = list(self.arraystructure[self.data_len])

        new_element[self.index] = deepcopy(value)
        self.arraystructure[self.data_len] = tuple(new_element)
        self.data_len += 1


class TrajectoryGraph:
    def __init__(self, system):
        """
        Stores and analyzes the information of a kinetic MC trajectory
        system: system
        """

        self.node_count = len(system.get_states())

        self.graph = nx.DiGraph()

        for inode, state in enumerate(system.get_states()):
            # print('add_node', inode, state.label, state)

            self.graph.add_node(inode,
                                coordinates=[list(state.get_coordinates())],
                                state=state.label,
                                time=[0],
                                event_time=0,
                                index=[id(state)],
                                finished=False,
                                )

        self.supercell = np.array(system.supercell)
        self.system = system

        self.n_dim = len(system.molecules[0].get_coordinates())
        self.n_excitons = len(system.get_states())
        self.labels = {}
        self.times = [0]

        self.states = set()
        ce = {}
        for state in self.system.get_states():
            self.states.add(state.label)
            count_keys_dict(ce, state.label)

        self.current_excitons = [ce]

        # Calculations
        self._diff_coefficient = {}
        self._diff_tensor = {}
        self._diff_length_square = {}
        self._diff_length_tensor = {}
        self._coordinates_dict = {}
        self._times_dict = {}

    def _reset_calculation_data(self):
        self._diff_coefficient = {}
        self._diff_tensor = {}
        self._diff_length_square = {}
        self._diff_length_tensor = {}
        self._vector_list_info = {}
        self._coordinates_dict = {}
        self._times_dict = {}

    def _finish_node(self, inode):

        node = self.graph.nodes[inode]
        if not node['finished']:
            # index = change_step['donor']
            node['index'].append(node['index'][-1])
            node['coordinates'].append(node['coordinates'][-1])
            node['time'].append(self.times[-1] - node['event_time'])
            node['finished'] = True

    def _add_node(self, from_node, new_on_state, process_label=None):

        #if self.system.molecules[new_on_molecule].set_state(_GS_):
        #    print('Error in state: ', self.system.molecules[new_on_molecule].state.label)

        self.graph.add_edge(from_node, self.node_count, process_label=process_label)
        self.graph.add_node(self.node_count,
                            coordinates=[list(new_on_state.get_coordinates_absolute())],
                            state=new_on_state.label,
                            time=[0],
                            event_time=self.times[-1],
                            index=[id(new_on_state)],
                            finished=False
                            )
        self.node_count += 1
        return self.node_count - 1

    def _append_to_node(self, on_node, add_state):
        node = self.graph.nodes[on_node]

        node['index'].append(id(add_state))
        node['coordinates'].append(list(add_state.get_coordinates_absolute()))
        node['time'].append(self.times[-1] - node['event_time'])

    def add_step(self, process, time_step):
        """
        Adds trajectory step

        :param process: process occurred during time_step: {donor, process, acceptor}
        :param time_step: duration of the chosen process
        """

        # print('-------------------------------')
        # print('process: ', process.description)

        self.times.append(self.times[-1] + time_step)

        end_points = [node for node in self.graph.nodes
                      if len(list(self.graph.successors(node))) == 0 and not self.graph.nodes[node]['finished']]

        # print('end_points', end_points)
        # print('initial: ', process.initial)
        node_links = {}
        created_nodes = {}
        for state in process.initial:
            for inode in end_points:
                node = self.graph.nodes[inode]
                if node['index'][-1] == id(state):
                    node_links[state] = inode
                    break

        # print('node_links', node_links)
        if len(node_links) == 0:
            exit()

        for initial_state, inode in node_links.items():

            if initial_state in process.get_transport_connections():
                for final_state in process.get_transport_connections()[initial_state]:
                    # Transfer
                    # print('append:', inode, final_state.label, final_state)
                    self._append_to_node(on_node=inode,
                                         add_state=final_state)
            else:
                # splitting & merging
                self._finish_node(inode)

                for final_state in process.get_transition_connections()[initial_state]:
                    if final_state in created_nodes:
                        # print('add edge: ', inode, '-', created_nodes[final_state], final_state.label, final_state)
                        self.graph.add_edge(inode,
                                            created_nodes[final_state],
                                            process_label=process.description)
                    else:
                        # print('add_node: ', inode, '-', self.node_count, final_state.label, final_state)
                        created_nodes[final_state] = self._add_node(from_node=inode,
                                                                    new_on_state=final_state,
                                                                    process_label=process.description)

        # print('------', initial_state, [state.label for state in process.initial], [state.label for state in process.final])
        ce = {}
        for state in self.system.get_states():
            self.states.add(state.label)
            count_keys_dict(ce, state.label)

        self.current_excitons.append(ce)

        # Reset data
        self._reset_calculation_data()

    def _start_point_nodes(self):

        #print()
        #print('edges', list(self.graph.nodes))
        start_points = []
        #transposed_edges_list = list(zip(*self.graph.out_edges))
        for node in self.graph.nodes:
            if len(self.graph.in_edges(node)) == 0:
                start_points.append(node)

        return start_points

    def plot_graph(self):

        # cmap = cm.get_cmap('Spectral')
        # default matplotlib color cycle list
        color_list = [u'#1f77b4', u'#ff7f0e', u'#2ca02c', u'#d62728', u'#9467bd', u'#8c564b', u'#e377c2', u'#7f7f7f', u'#bcbd22', u'#17becf']

        colors_map = {}
        node_map = {}
        for i, state in enumerate(self.get_states()):
            colors_map[state] = np.roll(color_list, -i)[0]
            node_map[state] = []

        for node in self.graph:
            state = self.graph.nodes[node]['state']
            node_map[state].append(node)

        # pos = nx.spring_layout(self.graph)
        pos = nx.drawing.nx_agraph.graphviz_layout(self.graph, prog='dot')
        for state in self.get_states():
            nx.draw_networkx_nodes(self.graph,
                                   pos=pos,
                                   nodelist=node_map[state],
                                   node_color=colors_map[state],
                                   label=state)
        nx.draw_networkx_edges(self.graph, pos=pos)
        nx.draw_networkx_labels(self.graph, pos=pos)
        plt.legend()

        return plt

    def get_states(self):
        return self.states

    def get_dimension(self):
        return self.n_dim

    def get_graph(self):
        return self.graph

    def get_simulation_times(self):
        return self.times

    def get_times(self, state=None):

        if state in self._times_dict:
            return self._times_dict[state]

        if state is None:
            self._times_dict[state] = self._vector_list_global()[1]
        else:
            self._times_dict[state] = self._vector_list(state)[1]

        return self._times_dict[state]

    def get_coordinates(self, state=None):

        if state in self._coordinates_dict:
            return self._coordinates_dict[state]

        if state is None:
            self._coordinates_dict[state] = np.array([np.array(ar).T.tolist() for ar in self._vector_list_global()[0]])
            #print('--> none shape', np.array(self._coordinates_dict[state]).shape)
            #print(self._coordinates_dict[state][0])
            #print(self._coordinates_dict[state][1])

            #exit()
        else:
            self._coordinates_dict[state] = [np.array(ar).T.tolist() for ar in self._vector_list(state)[0]]
            #print('--> state shape', np.array(self._coordinates_dict[state]).shape)
            #print(self._coordinates_dict[state][0])
            #print(self._coordinates_dict[state][1])

        return self._coordinates_dict[state]

    def get_distances(self):
        return [np.linalg.norm(coor, axis=1).tolist() for coor in self.get_coordinates()]
        #return np.linalg.norm(self.get_coordinates(), axis=2).tolist()

    def _vector_list(self, state):

        node_list = [node for node in self.graph.nodes if self.graph.nodes[node]['state'] == state]

        vector = []
        times = []
        for node in node_list:

            if len(self.graph.nodes[node]['time']) == 1:
                continue

            times.append(self.graph.nodes[node]['time'])
            #times.append(self.graph.nodes[node]['time'][:-1])

            initial = np.array(self.graph.nodes[node]['coordinates'][0])
            vector.append(np.array([np.array(coordinate) - initial for coordinate in self.graph.nodes[node]['coordinates']]).T.tolist())
            #vector.append(np.array([np.array(coordinate) - initial for coordinate in self.graph.nodes[node]['coordinates'][:-1]]).T.tolist())


        return [vector, times]

    def _vector_list_global(self):

        vector_list = []
        times_list = []
        for node_list in self._trajectory_node_path():

            vector = [[0.0] * self.get_dimension()]
            times = [0]
            t_ini = self.graph.nodes[0]['time'][0]
            coor_ini = np.array([0.0] * self.get_dimension())

            for node in node_list:
                #print('t_ini', t_ini)
                times += list(np.array(self.graph.nodes[node]['time'][1:]) + t_ini)
                initial = np.array(self.graph.nodes[node]['coordinates'][0])
                #print('v ', [np.array(coordinate) - initial for coordinate in self.graph.nodes[node]['coordinates']])
                vector += [np.array(coordinate) - initial - np.array(coor_ini) for coordinate in self.graph.nodes[node]['coordinates'][1:]]
                t_ini = times[-1]
                coor_ini = vector[-1]

            vector = np.array(vector).T.tolist()

            if len(times) == 1:
                continue

            vector_list.append(vector)
            times_list.append(times)

        return vector_list, times_list

    def _trajectory_node_path(self):

        trajectory_path = []
        for ini_node in self._start_point_nodes():
            same_initial_list = []

            def _recursive(inode, internal_list):
                end_of_trajectory = True
                for e in self.graph.out_edges(inode):
                    if e[0] == inode:
                        end_of_trajectory = False
                        internal_list = list(internal_list)
                        internal_list.append(e[1])
                        _recursive(e[1], tuple(internal_list))
                        internal_list = internal_list[:-1]
                if end_of_trajectory:
                    internal_list = list(internal_list)
                    same_initial_list.append(tuple(internal_list))

            _recursive(ini_node, [ini_node])
            trajectory_path += same_initial_list

        return trajectory_path

    def get_n_subtrajectories(self, state=None):
        return len(self.get_times(state))

    def get_n_segments(self, state=None):
        # get number of subtrajectories

        if state is None:
            state_list = self.get_states()
        else:
            state_list = [state]

        n_segments = 0
        for s in state_list:
            node_list = [node for node in self.graph.nodes if self.graph.nodes[node]['state'] == s]
            n_segments += len(node_list)
        return n_segments

    def get_diffusion(self, state):
        """
        Return the average diffusion coefficient defined as:

        DiffCoeff = 1/(2*z) * <DiffLen^2>/<time>

        :return: the diffusion coefficient
        """

        if state in self._diff_coefficient:
            return self._diff_coefficient[state]

        #print('times', self.get_times(state))
        #print('coord', self.get_coordinates(state))
        #print('len', len(self.get_times(state)), len(self.get_coordinates(state)))
        #print('sp:', self._start_point_nodes())

        slope_list = []
        for v, t in zip(self.get_coordinates(state), self.get_times(state)):

            vector2 = np.linalg.norm(v, axis=1)**2  # emulate dot product in axis 0
            #vector2 = np.diag(np.dot(vector.T, vector))

            #plt.plot(t, vector2, 'o', label=state)
            #plt.show()
            with np.errstate(invalid='ignore'):
                slope, intercept, r_value, p_value, std_err = stats.linregress(t, vector2)

            slope_list.append(slope)

        # plt.show()
        if len(slope_list) == 0:
            self._diff_coefficient[state] = None
        else:
            self._diff_coefficient[state] = np.mean(slope_list)/(2 * self.get_dimension())

        return self._diff_coefficient[state]

    def get_diffusion_tensor(self, state):

        if state in self._diff_tensor:
            return self._diff_tensor[state]

        tensor_x_list = []
        for v, t in zip(self.get_coordinates(state), self.get_times(state)):
            vector = np.array(v).T

            tensor_x = []
            for v1 in vector:
                tensor_y = []
                for v2 in vector:
                    #print('v1, v2:', v1, v2)
                    vector2 = np.multiply(v1, v2)
                    with np.errstate(invalid='ignore'):
                        slope, intercept, r_value, p_value, std_err = stats.linregress(t, vector2)
                    tensor_y.append(slope)
                tensor_x.append(tensor_y)

            tensor_x_list.append(np.array(tensor_x))

        if len(tensor_x_list) == 0:
            self._diff_tensor[state] = None
        else:
            self._diff_tensor[state] = np.mean(tensor_x_list, axis=0) / 2

        return self._diff_tensor[state]

    def get_number_of_cumulative_excitons(self, state=None):
        time = []
        node_count = []
        print('This is wrong!!, accumulated')
        for node in self.graph.nodes:
            time.append(self.graph.nodes[node]['event_time'])
            if state is not None:
                if self.graph.nodes[node]['event_time'] == state:
                    node_count.append(node_count[-1]+1)
            else:
                node_count.append(node)
        return time, node_count

    def get_number_of_excitons(self, state=None):
        excitations_count = []
        for t, status in zip(self.times, self.current_excitons):
            if state is None:
                excitations_count.append(np.sum(list(status.values())))
            else:
                if state in status:
                    excitations_count.append(status[state])
                else:
                    excitations_count.append(0)

        return excitations_count

    def plot_number_of_cumulative_excitons(self, state=None):
        t, n = self.get_number_of_cumulative_excitons(state)
        plt.plot(t, n, '-o')
        return plt

    def plot_number_of_excitons(self, state=None):
        n = self.get_number_of_excitons(state)
        plt.plot(self.times, n, '-o')
        return plt

    def get_number_of_nodes(self):
        return self.graph.number_of_nodes()

    def plot_2d(self, state=None, supercell_only=False, show_warnings=True):

        if state is None:
            node_list = [node for node in self.graph.nodes]
        else:
            node_list = [node for node in self.graph.nodes if self.graph.nodes[node]['state'] == state]

        if len(node_list) == 0 and show_warnings:
            warnings.warn('Exciton {} not found for plot'.format(state))

        t = []
        coordinates = []
        for node in node_list:
            t += [self.graph.nodes[node]['time'] for node in node_list]

            if supercell_only:
                coordinates += [self.graph.nodes[node]['coordinates'] for node in node_list]

            else:
                vector = []
                for coordinate in self.graph.nodes[node]['coordinates']:
                    vector.append(np.array(coordinate))
                coordinates += vector
                # print(vector)
                plt.plot(np.array(vector).T[0], np.array(vector).T[1], '-o')
                # plt.show()

        if self.get_dimension() != 2:
            raise Exception('plot_2d can only be used in 2D systems')

        coordinates = np.array(coordinates).T

        if len(coordinates) == 0:
            # warnings.warn('No data for state {}'.format(state))
            return plt

        # plt.plot(coordinates[0], coordinates[1], '-o')
        plt.title('exciton trajectories ({})'.format(state if state is not None else 'All'))

        return plt

    def get_distances_vs_times(self, state=None):

        if state is None:
            node_list = [node for node in self.graph.nodes]
        else:
            node_list = [node for node in self.graph.nodes if self.graph.nodes[node]['state'] == state]

        t = []
        coordinates = []
        for node in node_list:
            t += self.graph.nodes[node]['time']

            vector = []
            initial = self.graph.nodes[node]['coordinates'][0]

            for coordinate in self.graph.nodes[node]['coordinates']:
                vector.append(np.array(coordinate) - initial)

            # print('->', [np.linalg.norm(v, axis=0) for v in vector])
            # print('->', t)
            # plt.plot(self.graph.nodes[node]['time'], [np.linalg.norm(v, axis=0) for v in vector], '-o')

            coordinates += vector

        vector = np.array(coordinates).T

        if len(coordinates) == 0:
            # warnings.warn('No data for state {}'.format(state))
            return [], []

        vector = np.linalg.norm(vector, axis=0)

        return vector, t

    def get_max_distances_vs_times(self, state):

        if state is None:
            node_list = [node for node in self.graph.nodes]
        else:
            node_list = [node for node in self.graph.nodes if self.graph.nodes[node]['state'] == state]

        t = []
        coordinates = []
        for node in node_list:
            t += self.graph.nodes[node]['time']

            vector = []
            initial = np.array(self.graph.nodes[node]['coordinates'][0])
            final = np.array(self.graph.nodes[node]['coordinates'][-1])
            vector.append(final - initial)

            coordinates += vector

        vector = np.array(coordinates).T

        if len(coordinates) == 0:
            # warnings.warn('No data for state {}'.format(state))
            return [], []

        vector = np.linalg.norm(vector, axis=0)

        return vector, t

    def plot_distances(self, state=None):

        vector, t = self.get_distances_vs_times(state)

        # print(t)
        plt.title('diffusion distances ({})'.format(state if state is not None else 'All'))
        plt.plot(t, vector, '.')
        plt.xlabel('Time (ns)')
        plt.ylabel('Distance (Angs)')

        return plt

    def get_lifetime(self, state):

        if len(self.get_times(state)) == 0:
            return None

        if state is None:
            return self.get_simulation_times()[-1]

        return np.mean([t[-1] for t in self.get_times(state)])

    def get_time_ratio(self, state):

        t_state = np.sum([t[-1] for t in self.get_times(state)])
        t_tot = np.sum([t[-1] for t in self.get_times(None)])

        return t_state/t_tot

    def get_diffusion_length_square(self, state):

        if state in self._diff_length_square:
            return self._diff_length_square[state]

        vector_final = [v[-1] for v in self.get_coordinates(state)]

        if len(vector_final) == 0:
            return None

        dot_list = np.linalg.norm(vector_final, axis=1)**2

        self._diff_length_square[state] = np.average(dot_list)
        return self._diff_length_square[state]

    def get_diffusion_length_square_tensor(self, state):

        if state in self._diff_length_tensor:
            return self._diff_length_tensor[state]

        vector_final = [v[-1] for v in self.get_coordinates(state)]

        tensor = []
        for vector in vector_final:
            tensor_x = []
            for v1 in vector:
                tensor_y = []
                for v2 in vector:
                    tensor_y.append(v1*v2)
                tensor_x.append(tensor_y)

            tensor.append(np.array(tensor_x))

        if len(tensor) == 0:
            return None

        self._diff_length_tensor[state] = np.average(tensor, axis=0)*self.get_dimension()
        return self._diff_length_tensor[state]


class TrajectoryGraph2(TrajectoryGraph):
    def __init__(self, system):
        """
        Stores and analyzes the information of a kinetic MC trajectory
        system: system
        """

        exit()

        self.node_count = len(system.get_states())

        self.graph = nx.DiGraph()

        if not os.path.exists('test_map'):
            os.mkdir('test_map')

        self.mapped_list = []
        for i, state in enumerate(system.get_states()):

            mem_array = np.require(np.memmap('test_map/array_{}_{}_{}'.format(id(self), os.getpid(), i),
                                             dtype=[('coordinates', object),
                                                    ('cell_state', object),
                                                    ('time', object),
                                                    ('index', object)],
                                             mode='w+', shape=(1,)),
                                   requirements=['O'])

            mem_array[:] = (
                            #state.get_center().get_coordinates(),
                            #state.get_center().cell_state,
                            0.0,
                            system.get_molecule_index(state.get_center()))

            self.graph.add_node(i,
                                coordinates=ArrayHandler(mem_array, 'coordinates'),
                                state=state.label,
                                cell_state=ArrayHandler(mem_array, 'cell_state'),
                                time=ArrayHandler(mem_array, 'time'),
                                event_time=0,
                                index=ArrayHandler(mem_array, 'index'),
                                finished=False,
                                )
            self.mapped_list.append((mem_array, 'test_map/array_{}_{}_{}'.format(id(self), os.getpid(), i)))

        self.supercell = system.supercell
        self.system = system

        self.n_dim = len(system.molecules[0].get_coordinates())
        self.n_excitons = len(system.get_states())
        self.labels = {}

        mem_array_t = np.require(np.memmap('test_map/array_{}_{}_{}'.format(id(self), os.getpid(), 't'),
                                           dtype=[('times', object)],
                                           mode='w+', shape=(1,)),
                                 requirements=['O'])

        mem_array_t[:] = (0)
        self.times = ArrayHandler(mem_array_t, 'times')
        self.mapped_list.append((mem_array_t, 'test_map/array_{}_{}_{}'.format(id(self), os.getpid(),'t')))

        self.states = set()
        ce = {}
        for state in self.system.get_states():
            self.states.add(state.label)
            count_keys_dict(ce, state.label)
        self.current_excitons = [ce]

    def _add_node(self, from_node, new_on_molecule, process_label=None):

        if self.system.molecules[new_on_molecule].set_state(_GS_):
            print('Error in state: ', self.system.molecules[new_on_molecule].state.label)
            exit()

        mem_array = np.require(np.memmap('test_map/array_{}_{}_{}'.format(id(self), os.getpid(), self.node_count),
                                         dtype=[('coordinates', object),
                                                ('cell_state', object),
                                                ('time', object),
                                                ('index', object)],
                                         mode='w+', shape=(1,)),
                               requirements=['O'])

        mem_array[:] = (self.system.molecules[new_on_molecule].get_coordinates(),
                        self.system.molecules[new_on_molecule].cell_state,
                        0.0,
                        new_on_molecule)

        self.graph.add_edge(from_node, self.node_count, process_label=process_label)
        self.graph.add_node(self.node_count,
                            coordinates=ArrayHandler(mem_array, 'coordinates'),
                            state=self.system.molecules[new_on_molecule].state.label,
                            cell_state=ArrayHandler(mem_array, 'cell_state'),
                            time=ArrayHandler(mem_array, 'time'),
                            event_time=self.times[-1],
                            index=ArrayHandler(mem_array, 'index'),
                            finished=False
                            )

        self.mapped_list.append((mem_array, 'test_map/array_{}_{}_{}'.format(id(self), os.getpid(), self.node_count)))

        self.node_count += 1

    def __del__(self):
        for mapped_array, filename in self.mapped_list:
            del mapped_array
            try:
                os.remove(filename)
            except FileNotFoundError:
                pass



import json
import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D

from kimonet.analysis.trajectory_graph import TrajectoryGraph as Trajectory
from kimonet.analysis.trajectory_analysis import TrajectoryAnalysis
from kimonet.system.state import ground_state as _GS_
import warnings


def visualize_system(system, dipole=None):

    ndim = system.molecules[0].get_dim()
    #fig, ax = plt.subplots()
    fig = plt.figure()

    fig.suptitle('Orientation' if dipole is None else 'TDM {}'.format(dipole))
    if ndim == 3:
        ax = fig.gca(projection='3d')
        ax.set_zlabel('Z')
    else:
        ax = fig.gca()

    ax.set_xlabel('X')
    ax.set_ylabel('Y')

    # plt.xlim([0, np.dot([1, 0], system.supercell[0])])
    # plt.ylim([0, np.dot([0, 1], system.supercell[1])])

    list_state_labels = np.unique([s.label for s in system.get_states()] + [_GS_.label])
    color_list = ['red', 'blue', 'green', 'orange', 'grey']

    colors = {l: np.roll(color_list, -i)[0] for i, l in enumerate(list_state_labels)}

    for i, molecule in enumerate(system.molecules):
        c = molecule.get_coordinates()
        if dipole is None:
            o = molecule.get_orientation_vector()
        else:
            o = molecule.get_transition_moment(to_state=dipole)

        if np.linalg.norm(o) == 0:
            continue

        if ndim == 1:
            ax.quiver(c[0], 0, o[0], 0, color=colors[molecule.state.label])
        if ndim == 2:
            ax.quiver(c[0], c[1], o[0], o[1], color=colors[molecule.state.label])
            ax.text(c[0], c[1], '{}'.format(i), fontsize=12)
        if ndim == 3:
            ax.quiver(c[0], c[1], c[2], o[0], o[1], o[2], normalize=True, length=5, color=colors[molecule.state.label])
            # ax.quiver(c[0], c[1], c[2], o[0], o[1], o[2], length=0.1, normalize=True)
            ax.text(c[0], c[1], c[2], '{}'.format(i), fontsize=12)

    # Plot lattice vectors
    if ndim > 1:
        for lattice_vector in system.supercell:
            ax.plot(*np.array([[0]*ndim, lattice_vector]).T)

    for i in range(ndim):
        for j in range(i+1, ndim):
            ax.plot(*np.array([system.supercell[i], system.supercell[i] + system.supercell[j]]).T, color='black')
            ax.plot(*np.array([system.supercell[j], system.supercell[j] + system.supercell[i]]).T, color='black')
            if ndim == 3:
                ax.plot(*np.array([system.supercell[i]+system.supercell[j],
                                   system.supercell[0]+system.supercell[1]+system.supercell[2]]).T, color='black')

    # ax.quiverkey(q, X=0.3, Y=1.1, U=10,
    #               label='Quiver key, length = 10', labelpos='E')

    plt.show()


def plot_polar_plot(tensor_full, plane=(0, 1), title='', max=None, crystal_labels=False):

    tensor = np.array(tensor_full)[np.array(plane)].T[np.array(plane)].T

    r = []
    theta = []
    for i in np.arange(0, np.pi*2, 0.01):
        unit_vector = np.array([np.cos(i), np.sin(i)])
        r.append(np.dot(unit_vector, np.dot(tensor, unit_vector)))  # <n|D|n>
        theta.append(i)

    if max is None:
        max = np.max(np.nan_to_num(r)) * 1.2

    labels = {'cartesian': ['x', 'y', 'z'],
              'crystal': ['a', 'b', 'c']}

    if crystal_labels:
        labels_plot = [labels['crystal'][i] for i in plane]
    else:
        labels_plot = [labels['cartesian'][i] for i in plane]

    fig = plt.figure()
    ax = fig.add_axes([0.1, 0.05, 0.8, 0.8], polar=True)
    #ax = plt.subplot(111, projection='polar')
    ax.arrow(0., 0., np.pi, max,  edgecolor='black', lw=1, zorder=5)
    ax.arrow(0., 0., 3./2*np.pi, max,  edgecolor='black', lw=1, zorder=5)
    ax.annotate("", xy=(0, max), xytext=(0, 0), arrowprops=dict(arrowstyle="->"))
    ax.annotate("", xy=(np.pi/2, max), xytext=(0, 0), arrowprops=dict(arrowstyle="->"))
    ax.plot(theta, r)
    #plt.polar(theta, r)
    ax.set_rmax(max)
    ax.set_rticks(list(np.linspace(0.0, max, 8)))  # Less radial ticks
    ax.set_rlabel_position(-22.5)  # Move radial labels away from plotted line
    ax.grid(True)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        ax.set_xticklabels(['{}'.format(labels_plot[0]), '', '{}'.format(labels_plot[1]), '', '', '', '', ''])

    ax.set_title(title, va='bottom')
    plt.show()



from kimonet.utils.units import HBAR_PLANCK, BOLTZMANN_CONSTANT
import numpy as np
from kimonet.system.vibrations import NoVibration
from scipy.integrate import quad
from copy import deepcopy
from kimonet.system.state import ground_state as _GS_

overlap_data = {}


def ordered_states(state_list):
    """
    get GS states behind and other states at front
    :param state_list: List of states behind and
    :return: ordered List
    """
    labels = [s.label for s in state_list]
    indices = np.argsort(labels)
    state_list = np.array(state_list)[indices].tolist()

    ordered_list = []
    for state in state_list:
        if state.label == _GS_.label:
            ordered_list.append(state)
        else:
            ordered_list.insert(0, state)

    return tuple(ordered_list)


class BaseProcess(object):
    def __init__(self,
                 initial_states,
                 final_states,
                 description='',
                 arguments=None,
                 do_copy=True,
                 ):

        if do_copy:
            initial_states = tuple([s.copy() for s in initial_states])
            final_states = tuple([s.copy() for s in final_states])

        self._initial = ordered_states(initial_states)
        self._final_test = ordered_states(deepcopy(final_states))

        self._final = None

        self.description = description
        self.cell_states = {}
        self.arguments = arguments if arguments is not None else {}
        self._supercell = None
        self._transition_connect = None
        self._transport_connect = None
        self._is_symmetry = None
        self._cell_increment = None

        # Check input coherence
        total_size_initial = np.sum([state.size for state in initial_states])
        total_size_final = np.sum([state.size for state in final_states])

        # Check initial & final states sizes match
        assert total_size_initial == total_size_final

    #    def __str__(self):
#        return 'donor/acceptor : {} {}\n'.format(self.donor.state, self.acceptor.state) \
#               + 'initial : {} {}\n'.format(self.initial[0], self.initial[1]) \
#               + 'final : {} {}\n'.format(self.final[0], self.final[1])

    @property
    def final(self):
        if self._final is None:
            self._final = deepcopy(self.final_test)
            for state in self._final:
                for mol in state.get_molecules():
                    mol.cell_state = self.cell_states[mol]
                    mol.set_state(state)

        return self._final

    @property
    def initial_absolute(self):
        if self._cell_increment is None:
            return self.initial

        initial_copy = deepcopy(self.initial)
        for state in initial_copy[1:]:
            for mol in state.get_molecules():
                mol.cell_state = self.cell_states[mol]
                mol.set_state(state)

        initial_copy[1].cell_state = np.array(initial_copy[0].cell_state) - np.array(self._cell_increment)
        return initial_copy

    def set_cell_increment(self, cell_increment):
        self._cell_increment = cell_increment

    @property
    def initial(self):
        return self._initial

    @initial.setter
    def initial(self, state_list):
        self._transition_connect = None
        self._transport_connect = None
        self._initial = state_list

    @property
    def final_test(self):
        return self._final_test

    @final_test.setter
    def final_test(self, state_list):
        self._transition_connect = None
        self._transport_connect = None
        self._final_test = state_list

    @property
    def supercell(self):
        if self._supercell is None:
            raise Exception('No supercell')
        return self._supercell

    @supercell.setter
    def supercell(self, cell):
        self._supercell = cell
        for state in self.final_test:
            state.supercell = cell

    def is_symmetry(self):
        if self._is_symmetry is None:
            label_list = [s.label for s in self._initial]
            self._is_symmetry = not len(np.unique(label_list)) == len(label_list)

        return self._is_symmetry

    def get_self_interaction_process(self):
        """
        By default not allow self interaction in this process
        :return:
        """
        return None

    def reset_cell_states(self):
        self.cell_states.clear()
        for state in self.final_test:
            for mol in state.get_molecules():
                self.cell_states[mol] = np.zeros(mol.get_dim())

    def get_molecules(self):

        molecules_list = []
        for state in self._initial:
            molecules_list += state.get_molecules()

        return molecules_list

    def get_transport_connections(self):
        """
        Get the connections between initial and final states
        that are considered to be transport (same state moving)
        :return:
        """
        if self._transport_connect is None:
            self._transport_connect = {}
            for istate in self._initial:
                for fstate in self._final_test:
                    if istate.label == fstate.label and istate.label != _GS_.label:
                        if istate in self._transport_connect:
                            self._transport_connect[istate].append(fstate)
                        else:
                            self._transport_connect[istate] = [fstate]

        return self._transport_connect

    def get_transition_connections(self):
        """
        Get the connections between initial and final states
        that are considered to be transitions (state converting to other)
        :return:
        """

        if self._transition_connect is None:
            self._transition_connect = {}
            inital_states = [s for s in self._initial if s.label != _GS_.label]
            final_states = [s for s in self._final_test if s.label != _GS_.label]
            if len(inital_states) == 1:
                self._transition_connect[inital_states[0]] = []
                for fstate in final_states:
                    self._transition_connect[inital_states[0]].append(fstate)
            elif len(final_states) == 1:
                for istate in inital_states:
                    self._transition_connect[istate] = [final_states[0]]
            else:
                for istate in self._initial:
                    self._transition_connect[istate] = []
                    for fstate in final_states:
                        self._transition_connect[istate].append(fstate)

        return self._transition_connect


class GoldenRule(BaseProcess):
    def __init__(self,
                 initial_states,
                 final_states,
                 electronic_coupling_function,
                 description='',
                 arguments=None,
                 vibrations=NoVibration(),
                 ):

        self._coupling_function = electronic_coupling_function
        self._vibrations = vibrations
        BaseProcess.__init__(self, initial_states, final_states, description, arguments)

    @property
    def vibrations(self):
        return self._vibrations

    def get_fcwd(self):
        transition_donor = (self.initial[0], self.final[1])
        transition_acceptor = (self.initial[1], self.final[0])

        donor_vib_dos = self.vibrations.get_vib_spectrum(*transition_donor)  # (transition_donor)
        acceptor_vib_dos = self.vibrations.get_vib_spectrum(*transition_acceptor)  # (transition_acceptor)

        # print(donor_vib_dos)
        info = str(hash(donor_vib_dos) + hash(acceptor_vib_dos))

        # the memory is used if the overlap has been already computed
        if info in overlap_data:
            return overlap_data[info]

        def overlap(x):
            return donor_vib_dos(x) * acceptor_vib_dos(x)

        d = BOLTZMANN_CONSTANT*1000*10
        sup_range = np.max([self.initial[0].energy, self.initial[0].energy]) + d
        inf_range = np.min([self.initial[0].energy, self.initial[0].energy]) - d


        if isinstance(self.vibrations, NoVibration):
            # Treat donor_vib_dos & acceptor_vib_dos as delta functions
            overlap_data[info] = 1.0 if donor_vib_dos - acceptor_vib_dos == 0 else 0.0
        else:
            overlap_data[info] = quad(overlap, inf_range, sup_range, epsabs=1e-5, limit=1000)[0]

        return overlap_data[info]

    def get_electronic_coupling(self):
        return self._coupling_function(self.initial, self.final, **self.arguments)

    def get_rate_constant(self):
        e_coupling = self.get_electronic_coupling()
        # spectral_overlap = general_fcwd(self.donor, self.acceptor, self, conditions)

        spectral_overlap = self.get_fcwd()

        return 2 * np.pi / HBAR_PLANCK * e_coupling ** 2 * spectral_overlap  # Fermi's Golden Rule


class DirectRate(BaseProcess):
    def __init__(self,
                 initial_states,
                 final_states,
                 rate_constant_function,
                 description='',
                 arguments=None
                 ):

        self.rate_function = rate_constant_function
        BaseProcess.__init__(self, initial_states, final_states, description, arguments)

    def get_rate_constant(self):
        return self.rate_function(self.initial, self.final, **self.arguments)


class SimpleRate(BaseProcess):
    def __init__(self,
                 initial_states,
                 final_states,
                 rate_constant,
                 description='',
                 arguments=None
                 ):

        self._rate_constant = rate_constant
        BaseProcess.__init__(self, initial_states, final_states, description, arguments)

    def get_rate_constant(self):
        return self._rate_constant


class SimpleRateBounded(BaseProcess):
    def __init__(self,
                 initial_states,
                 final_states,
                 rate_constant,
                 range_distance,
                 step_param=1.0,
                 description='',
                 arguments=None
                 ):
        """
        :param initial_states:
        :param final_states:
        :param rate_constant:
        :param range_distance: max
        :param step_param: derivative of sigmoid function at range_distance
        :param description:
        :param arguments:
        """

        self._range_distance = range_distance
        self._rate_constant = rate_constant
        self._step_param = step_param
        BaseProcess.__init__(self, initial_states, final_states, description, arguments)

    def get_rate_constant(self):
        r_vector = np.array(self.initial_absolute[0].get_coordinates_absolute()) - \
                   np.array(self.initial_absolute[1].get_coordinates_absolute())
        distance = np.linalg.norm(r_vector)

        def sigmoid_function(x, distance, step_param=1.0):
            return 1.0 / (1.0 + np.exp(1.0 * step_param * 4 * (x - distance)))

        return self._rate_constant * sigmoid_function(distance, self._range_distance, step_param=self._step_param)


class InternalConversion(BaseProcess):
    def __init__(self,
                 initial_state,
                 final_state,
                 coupling=None,
                 vibrations=NoVibration(),
                 k1=0,
                 description='',
                 arguments=None
                 ):

        self._k1 = k1
        self._coupling = coupling
        self._vibrations = vibrations

        BaseProcess.__init__(self, [initial_state], [final_state], description, arguments)

    def get_rate_constant(self):
        delta_g = self.final_test[0].energy - self.initial[0].energy

        return self._k1 + 2*np.pi/HBAR_PLANCK * self._coupling**2 * self.get_fcwd()

    @property
    def vibrations(self):
        return self._vibrations

    def get_fcwd(self):
        transition = (self.final[0], self.initial[0])

        vib_dos = self.vibrations.get_vib_spectrum(*transition)

        if isinstance(self.vibrations, NoVibration):
            return 1.0 if vib_dos == 0 else 0.0

        return vib_dos(0)


class DecayRate(BaseProcess):
    def __init__(self,
                 initial_state,
                 final_state,
                 decay_rate_function,
                 description='',
                 arguments=None
                 ):

        BaseProcess.__init__(self, [initial_state], [final_state], description, arguments)
        self.rate_function = decay_rate_function

    def get_rate_constant(self):
        return self.rate_function(self.initial, self.final, **self.arguments)



import numpy as np
from kimonet.utils.units import SPEED_OF_LIGHT, HBAR_PLANCK
from kimonet.core.processes.transitions import Transition
from kimonet.system.state import ground_state as _GS_
from kimonet.utils.units import ATOMIC_TO_ANGS_EL


# Decay functions
def einstein_radiative_decay(initial, final, g1=1, g2=1, transitions=None, transition_moment=None):
    """
    Einstein radiative decay

    :param molecule:
    :param g1: degeneracy of target state
    :param g2: degeneracy of origin state

    :return: decay rate constant
    """
    deexcitation_energy = initial[0].energy - final[0].energy

    mu = np.array(transitions[transitions.index(Transition(initial[0], final[0]))].tdm) * ATOMIC_TO_ANGS_EL

    # mu = np.array(transition_moment[Transition(initial[0], final[0])]) * ATOMIC_TO_ANGS_EL
    mu2 = np.dot(mu, mu)  # transition moment square norm.
    alpha = 1.0 / 137.036
    return float(g1)/g2 * alpha * 4 * deexcitation_energy ** 3 * mu2 / (3 * SPEED_OF_LIGHT ** 2 * HBAR_PLANCK ** 3)


def triplet_triplet_annihilation(molecule):
    f = 1
    ct = 1
    ptta = 50
    lifetime = 10
    a0 = 0.529177249
    t = ct*a0**-3
    return 1/(f*lifetime*t)*(1/ptta-1)**-1



from setuptools import setup, Extension
import numpy

include_dirs_numpy = [numpy.get_include()]


def get_version_number():
    main_ns = {}
    for line in open('kimonet/__init__.py', 'r').readlines():
        if not(line.find('__version__')):
            exec(line, main_ns)
            return main_ns['__version__']


def check_compiler():
    import subprocess
    output = subprocess.Popen(['gcc'], stderr=subprocess.PIPE).communicate()[1]
    if b'clang' in output:
        return 'clang'
    if b'gcc' in output:
        return 'gcc'


if check_compiler() == 'clang':
    forster = Extension('kimonet.core.processes.forster',
                        extra_compile_args=['-std=c99'],
                        include_dirs=include_dirs_numpy,
                        sources=['c/forster.c'])

else:
    print ('openmp is used')
    forster = Extension('kimonet.core.processes.forster',
                        extra_compile_args=['-std=c99', '-fopenmp'],
                        extra_link_args=['-lgomp'],
                        include_dirs=include_dirs_numpy,
                        sources=['c/forster.c'])

setup(name='kimonet',
      version=get_version_number(),
      description='kimonet module',
      author='Abel Carreras',
      url='https://github.com/abelcarreras/kimonet',
      author_email='abelcarreras83@gmail.com',
      packages=['kimonet',
                'kimonet.analysis',
                'kimonet.core',
                'kimonet.core.processes',
                'kimonet.system',
                'kimonet.utils'],
      install_requires=['numpy', 'scipy', 'matplotlib', 'networkx', 'h5py', 'pygraphviz'],
      license='MIT License',
      ext_modules=[forster])



import numpy as np
import itertools
import copy
from kimonet.utils import distance_vector_periodic
from kimonet.system.state import ground_state as _GS_
from kimonet.utils.combinations import get_molecules_centered_in_mol


class System(object):
    def __init__(self,
                 molecules,
                 supercell,
                 transfers=None,
                 decays=None,
                 cutoff_radius=10):

        self.molecules = molecules
        self.supercell = supercell
        self.neighbors = {}
        self.is_finished = False

        self._transfer_scheme = transfers if transfers is not None else []
        self._decay_scheme = decays if decays is not None else []

        self._cutoff_radius = cutoff_radius

        # search for states
        self._states = []
        for molecule in self.molecules:
            if molecule.state.label != _GS_.label:
                self._states.append(molecule.state)

        # speed up purpose savings
        self._gs_list = None
        self._state_neighbors = {}

        # set supercell to all states in the system
        for molecule in self.molecules:
            molecule.state.supercell = supercell

    @property
    def transfer_scheme(self):
        return self._transfer_scheme

    @transfer_scheme.setter
    def transfer_scheme(self, transfers):
        for transfer in transfers:
            transfer.supercell = self.supercell
        self._transfer_scheme = transfers

    @property
    def decay_scheme(self):
        return self._decay_scheme

    @decay_scheme.setter
    def decay_scheme(self, decays):
        for decay in decays:
            decay.supercell = self.supercell
        self._decay_scheme = decays

    @property
    def process_scheme(self):
        return self._decay_scheme + self._transfer_scheme

    @process_scheme.setter
    def process_scheme(self, processes):
        for process in processes:
            process.supercell = self.supercell
            if len(process.initial) == 1:
                self._decay_scheme.append(process)
            elif len(process.initial) == 2:
                self._transfer_scheme.append(process)
            else:
                raise Exception('Only processes involving 1 or 2 states are currently supported')

    def _reset_data(self):
        self._gs_list = None
        self._state_neighbors = {}


    def get_states(self):
        return self._states

    @property
    def cutoff_radius(self):
        return self._cutoff_radius

    @cutoff_radius.setter
    def cutoff_radius(self, cutoff):
        self._cutoff_radius = cutoff
        self._reset_data()

    def get_neighbours_num(self, center):

        radius = self.cutoff_radius
        center_position = self.molecules[center].get_coordinates()

        def get_supercell_increments(supercell, radius):
            # TODO: This function can be optimized as a function of the particular molecule coordinates
            v = np.array(radius/np.linalg.norm(supercell, axis=1), dtype=int) + 1  # here extensive approximation
            return list(itertools.product(*[range(-i, i+1) for i in v]))

        cell_increments = get_supercell_increments(self.supercell, radius)

        if not '{}_{}'.format(center, radius) in self.neighbors:
            neighbours = []
            jumps = []
            for i, molecule in enumerate(self.molecules):
                coordinates = molecule.get_coordinates()
                for cell_increment in cell_increments:
                    r_vec = distance_vector_periodic(coordinates - center_position, self.supercell, cell_increment)
                    if 0 < np.linalg.norm(r_vec) < radius:
                        neighbours.append(i)
                        jumps.append(cell_increment)

            neighbours = np.array(neighbours)
            jumps = np.array(jumps)

            self.neighbors['{}_{}'.format(center, radius)] = [neighbours, jumps]

        return self.neighbors['{}_{}'.format(center, radius)]

    def get_neighbours(self, ref_mol):

        radius = self.cutoff_radius
        center_position = ref_mol.get_coordinates()

        def get_supercell_increments(supercell, radius):
            # TODO: This function can be optimized as a function of the particular molecule coordinates
            v = np.array(radius/np.linalg.norm(supercell, axis=1), dtype=int) + 1  # here extensive approximation
            return list(itertools.product(*[range(-i, i+1) for i in v]))

        cell_increments = get_supercell_increments(self.supercell, radius)

        if not '{}_{}'.format(ref_mol, radius) in self.neighbors:
            neighbours = []
            jumps = []
            for molecule in self.molecules:
                coordinates = molecule.get_coordinates()
                for cell_increment in cell_increments:
                    r_vec = distance_vector_periodic(coordinates - center_position, self.supercell, cell_increment)
                    if 0 < np.linalg.norm(r_vec) < radius:
                        neighbours.append(molecule)
                        jumps.append(list(cell_increment))

            #jumps = np.array(jumps)

            self.neighbors['{}_{}'.format(ref_mol, radius)] = [neighbours, jumps]

        return self.neighbors['{}_{}'.format(ref_mol, radius)]

    def get_state_neighbors_copy(self, ref_state):
        """
        Not used for now. To be deprecated in the future

        :param ref_state:
        :return:
        """

        radius = self.cutoff_radius
        center_position = ref_state.get_coordinates_relative(self.supercell)

        def get_supercell_increments(supercell, radius):
            # TODO: This function can be optimized as a function of the particular molecule coordinates
            v = np.array(radius/np.linalg.norm(supercell, axis=1), dtype=int) + 1  # here extensive approximation
            return list(itertools.product(*[range(-i, i+1) for i in v]))

        cell_increments = get_supercell_increments(self.supercell, radius)

        neighbours = []
        for state in self.get_ground_states():
            coordinates = state.get_coordinates()
            for cell_increment in cell_increments:
                r_vec = distance_vector_periodic(coordinates - center_position, self.supercell, cell_increment)
                if 0 < np.linalg.norm(r_vec) < radius:
                    state = state.copy()
                    state.cell_state = ref_state.cell_state + cell_increment
                    neighbours.append(state)

        return neighbours

    def get_state_neighbors(self, ref_state):

        if ref_state not in self._state_neighbors:

            radius = self.cutoff_radius
            center_position = ref_state.get_coordinates_relative(self.supercell)

            def get_supercell_increments(supercell, radius):
                # TODO: This function can be optimized as a function of the particular molecule coordinates
                v = np.array(radius/np.linalg.norm(supercell, axis=1), dtype=int) + 1  # here extensive approximation
                return list(itertools.product(*[range(-i, i+1) for i in v]))

            cell_increments = get_supercell_increments(self.supercell, radius)

            # Get neighbor non ground states
            state_neighbors = []
            state_cell_incr = []
            # for state in self.get_ground_states() + self.get_states():
            for state in self.get_states():
                coordinates = state.get_coordinates()
                for cell_increment in cell_increments:
                    r_vec = distance_vector_periodic(coordinates - center_position, self.supercell, cell_increment)
                    if 0 < np.linalg.norm(r_vec) < radius:
                        state_neighbors.append(state)
                        state_cell_incr.append(list(cell_increment))

            # Include neighbor ground states
            state_neighbors_gs, state_cell_incr_gs = self.get_ground_states_improved(ref_state)
            state_cell_incr = state_cell_incr_gs + state_cell_incr
            state_neighbors = state_neighbors_gs + state_neighbors

            self._state_neighbors[ref_state] = [state_neighbors, state_cell_incr]

        return self._state_neighbors[ref_state]

    def get_ground_states(self):
        if self._gs_list is None:
            self._gs_list = []
            for mol in self.molecules:
                if (mol.state not in self._gs_list) and mol.state.label == _GS_.label:
                    self._gs_list.append(mol.state)

        return self._gs_list

    def get_ground_states_improved(self, ref_state):

        mol_list, cell_incr_list = self.get_neighbours(ref_state.get_center())

        gs_list = []
        gs_cell = []
        for mol, cell_incr in zip(mol_list, cell_incr_list):
            if mol.state.label == _GS_.label:
                gs_list.append(mol.state)
                gs_cell.append(cell_incr)

        return gs_list, gs_cell

    def get_molecule_index(self, molecule):
        return self.molecules.index(molecule)

    def reset(self):
        for molecule in self.molecules:
            molecule.set_state(_GS_.copy())
            molecule.state.supercell = self.supercell
            molecule.cell_state = np.zeros(molecule.get_dim())
        # self._centers = []
        self._reset_data()
        self.is_finished = False


    def copy(self):
        return copy.deepcopy(self)

    def get_num_molecules(self):
        return len(self.molecules)

    def get_number_of_excitations(self):
        return len(self._states)

    def add_excitation_index(self, state, index):

        state = state.copy()
        state.supercell = self.supercell

        gs_list = [s.get_center() for s in self.get_ground_states()]
        mol_list = get_molecules_centered_in_mol(self.molecules[index], gs_list, self.supercell,
                                                 size=state.size,
                                                 connected_distance=state.connected_distance)
        if mol_list is None:
            raise Exception('Not enough space is system')

        if state.label == _GS_.label:
            self._states.remove(self.molecules[index].state)
        else:
            self._states.append(state)

        for molecule in mol_list:
            molecule.set_state(state)

        self._reset_data()

    def remove_exciton(self, exciton):
        if exciton.label != _GS_.label:
            self._states.remove(exciton)
            for mol in exciton.get_molecules():
                mol.set_state(_GS_.copy())
                mol.state.supercell = self.supercell
            # if exciton in self._states:
#        else:
#            exciton.cell_state *= 0
#            exciton.reset_molecules()

        self._reset_data()

    def add_exciton(self, exciton):
        if exciton.label != _GS_.label:
            for mol in exciton.get_molecules():
                mol.set_state(exciton)

            self._states.append(exciton)
        else:
            exciton.cell_state *= 0
            exciton.reset_molecules()

        self._reset_data()

    def add_excitation_random(self, exciton_ref, n_states):
        max_cycles = len(self.molecules) * 2
        for i in range(n_states):
            for nc in range(max_cycles):

                gs_list = [s.get_center() for s in self.get_ground_states()]
                choice_molecule = np.random.choice(gs_list, 1)[0]

                exciton = exciton_ref.copy()
                exciton.supercell = self.supercell

                mol_list = get_molecules_centered_in_mol(choice_molecule, gs_list, self.supercell,
                                                         size=exciton.size,
                                                         connected_distance=exciton_ref.connected_distance)

                if mol_list is None:
                    if nc == max_cycles - 1:
                        raise Exception('Not enough space in system')
                    continue

                self._states.append(exciton)

                for molecule in mol_list:
                    molecule.set_state(exciton)

                self._reset_data()
                break

    def get_volume(self):
        return np.abs(np.linalg.det(self.supercell))

    def update(self, process):
        for initial in process.initial:
            self.remove_exciton(initial)

        for final in process.final_test:
            for mol in final.get_molecules():
                if final.label != _GS_.label:
                    mol.cell_state = process.cell_states[mol]
                mol.set_state(final)
            self.add_exciton(final)

        process.reset_cell_states()
        self._reset_data()

    def print_status(self):
        for i, mol in enumerate(self.molecules):
            print(i, mol.state.label, mol.state, mol.state.cell_state, mol.name, mol)
        print('----')
        for state in self.get_states():
            print(state.label, state, state.get_molecules())


if __name__ == '__main__':
    from kimonet.system.state import State
    from kimonet.system.molecule import Molecule

    s1 = State(label='s1', energy=1.0, multiplicity=1)
    molecule = Molecule()

    molecule1 = molecule.copy()
    molecule1.set_coordinates([0])
    molecule1.name = 'TypeA'

    molecule2 = molecule.copy()
    molecule2.set_coordinates([1])
    molecule2.name = 'TypeB'

    molecule3 = molecule.copy()
    molecule3.set_coordinates([2])
    molecule3.name = 'TypeC'

    # setup system
    system = System(molecules=[molecule1, molecule2, molecule3],
                    supercell=[[3]])

    system.add_excitation_index(s1, 2)
    print(system.get_ground_states())

    s1 = system.get_states()[0]
    print(s1, s1.label)
    s_list = system.get_state_neighbors_copy(s1)
    for s in s_list:
        print(s.label, s.get_coordinates_absolute())
        # print(s.label, s.get_coordinates_absolute(system.supercell) - s1.get_center().get_coordinates())

    print('----')

    s_list, c_list = system.get_state_neighbors(s1)
    for s, c in zip(s_list, c_list):
        print(s.label, s.get_coordinates_absolute() - np.dot(np.array(system.supercell).T, c))
        # print(s.label, s.get_coordinates_absolute(system.supercell) - s1.get_center().get_coordinates())


    print('----')
    m_list, c_list = system.get_neighbours(s1.get_center())
    print('m_list', m_list)
    for m, c in zip(m_list, c_list):
        if m.state.label != 's1':
            # print(np.array(system.supercell).T, c, np.dot(np.array(system.supercell).T, c))
            print(m.state.label, m.state.get_coordinates() - np.dot(np.array(system.supercell).T, c), m.state.get_center().name)



import h5py
from kimonet.analysis.trajectory_graph import TrajectoryGraph
import pickle
import numpy as np


def store_trajectory_list(trajectory_list, filename):
    f = h5py.File(filename, 'w')

    for i, trajectory in enumerate(trajectory_list):
        grp = f.create_group('{}'.format(i))
        grp.create_dataset('graph', data=np.void(pickle.dumps(trajectory.graph)))
        grp.create_dataset('node_count', data=trajectory.node_count)
        grp.create_dataset('system', data=np.void(pickle.dumps(trajectory.system)))
        grp.create_dataset('n_dim', data=trajectory.n_dim)
        grp.create_dataset('times', data=trajectory.times)
        grp.create_dataset('states', data=np.void(pickle.dumps(trajectory.states)))
        grp.create_dataset('current_excitons', data=np.void(pickle.dumps(trajectory.current_excitons)))

    f.close()

    pass


def load_trajectory_list(filename):
    f = h5py.File(filename, 'r')

    trajectory_list = []
    for dataset in f:
        graph = pickle.loads(f[dataset]['graph'][()].tostring())
        node_count = f[dataset]['node_count'][()]
        times = f[dataset]['times'][()]
        ndim = f[dataset]['n_dim'][()]
        system = pickle.loads(f[dataset]['system'][()].tostring())
        states = pickle.loads(f[dataset]['states'][()].tostring())
        current_excitons = pickle.loads(f[dataset]['current_excitons'][()].tostring())

        trajectory = TrajectoryGraph(system)
        trajectory.graph = graph
        trajectory.node_count = node_count
        trajectory.states = states
        trajectory.current_excitons = current_excitons
        trajectory.supercell = np.array(system.supercell)
        trajectory.n_excitons = len(system.get_states())
        trajectory.times = times
        trajectory.n_dim = ndim

        trajectory_list.append(trajectory)

    return trajectory_list



import numpy as np
from kimonet.utils import rotate_vector
import copy
from kimonet.system.state import ground_state as _GS_


class Molecule(object):

    def __init__(self,
                 name=None,
                 state=_GS_.copy(),
                 vdw_radius=1.0,  # Angstrom
                 coordinates=(0,),  # Angstrom
                 orientation=(0, 0, 0),  # Rx, Ry, Rz (radians)
                 ):
        """
        :param states_energies: dictionary {'state': energy} (eV)
        :param state: string containing the current state
        :param coordinates: the coordinates vector of the molecule within the system (Angstrom)
        :param orientation: 3d unit vector containing the orientation angles of the molecule defined in radiants respect X, Y and Z axes.
        """

        # set state energies to vibrations
        #vibrations.set_state_energies(state_energies)

        # self._state = self._labels_to_state[state]
        self._state = state
        self._coordinates = np.array(coordinates)
        self.orientation = np.array(orientation)
        self._cell_state = np.zeros_like(coordinates, dtype=int)
        self._vdw_radius = vdw_radius
        self.name = name

        state.remove_molecules()
        state.add_molecule(self)

    def __hash__(self):
        return hash((
                     # self._state.label,
                     self.name,
                     self._coordinates.tobytes(),
                     self.orientation.tobytes()
                     # np.array2string(self._cell_state, precision=12)
                     )
                    )

    def __eq__(self, other):
        return hash(self) == hash(other)

    def get_vdw_radius(self):
        return self._vdw_radius

    def get_dim(self):
        return len(self._coordinates)

    def set_coordinates(self, coordinates):
        """
        sets the coordinates of the molecule
        :param coordinates: coordinate vector
        """
        self._coordinates = np.array(coordinates)
        self._cell_state = np.zeros_like(self._coordinates, dtype=int)

    def get_coordinates(self):
        """
        sets the molecule coordinates
        :return: Array with the molecular coordinates.
        """
        return self._coordinates

    def set_orientation(self, orientation):
        """
        sets the orientation angles
        :param orientation: the orientation angles
        """
        self.orientation = np.array(orientation)

    def molecular_orientation(self):
        """
        :return: Array with the molecular orientation angles
        """
        return self.orientation

    def get_state_energy(self):
        return self._state.energy

    def copy(self):
        """
        returns a deep copy of this molecule
        :return: a copy of molecule
        """
        return copy.deepcopy(self)

    def get_orientation_vector(self):
        """
        return a vector that indicates the main reference orientation axis of the molecule.
        All other vector properties of the molecule are defined respect the molecule orientation
        This vector does not define the orientation completely, just serves as visual reference
        :return:
        """
        return rotate_vector([1, 0, 0][:self.get_dim()], self.orientation)

    def set_state(self, state):
        state.add_molecule(self)
        self._state = state

    @property
    def state(self):
        return self._state

    @property
    def cell_state(self):
        return self._cell_state

    @cell_state.setter
    def cell_state(self, c_state):
        self._cell_state = np.array(c_state)

    @property
    def cell_state_2(self):
        return self._cell_state + self.state.cell_state
